# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

import datetime
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from uuid import uuid4

import pyarrow
from orso import DataFrame
from orso import converters
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx import config
from opteryx import utils
from opteryx.constants import QueryStatus
from opteryx.constants import ResultType
from opteryx.exceptions import InconsistentSchemaError
from opteryx.exceptions import InvalidCursorStateError
from opteryx.exceptions import MissingSqlStatement
from opteryx.exceptions import SqlError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import QueryStatistics
from opteryx.utils import sql

PROFILE_LOCATION = config.PROFILE_LOCATION


class Cursor(DataFrame):
    """
    This class inherits from the orso DataFrame library to provide features such as fetch.
    """

    def __init__(self, connection):
        """
        Initializes the Cursor object, setting the initial state and binding the connection.

        Parameters:
            connection: Connection object
                The database connection object.
        """
        self.arraysize = 1
        self._connection = connection
        self._query_planner = None
        self._collected_stats = None
        self._plan = None
        self._qid = str(uuid4())
        self._statistics = QueryStatistics(self._qid)
        self._query_status = QueryStatus._UNDEFINED
        self._result_type = ResultType._UNDEFINED
        self._rowcount = None
        self._description: Optional[Tuple[Tuple[Any, ...], ...]] = None
        self._owns_connection = False
        self._closed = False
        self._executed = False
        DataFrame.__init__(self, rows=[], schema=[])

    @property
    def id(self) -> str:
        """The unique internal reference for this query.

        Returns:
            The unique query identifier as a string.
        """
        return self._qid

    def _inner_execute(
        self,
        operation: str,
        params: Union[Iterable, Dict, None] = None,
        visibility_filters: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """
        Executes a single SQL operation within the current cursor.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable/Dictionary, optional
                Parameters for the SQL operation, defaults to None.
        Returns:
            Results of the query execution.
        """
        from opteryx import system_statistics
        from opteryx.managers.execution import execute
        from opteryx.planner import query_planner

        if not operation:  # pragma: no cover
            raise MissingSqlStatement("SQL provided was empty.")

        self._connection.context.history.append(
            (operation, True, datetime.datetime.now(datetime.UTC))
        )

        start = time.time_ns()
        try:
            plan = query_planner(
                operation=operation,
                parameters=params,
                visibility_filters=visibility_filters,
                connection=self._connection,
                qid=self.id,
                statistics=self._statistics,
            )
        except RuntimeError as err:  # pragma: no cover
            raise SqlError(f"Error Executing SQL Statement ({err})") from err
        finally:
            self._statistics.time_planning += time.time_ns() - start

        results = execute(plan, statistics=self._statistics)
        system_statistics.queries_executed += 1

        if results is not None:
            # we can't update tuples directly
            entry = list(self._connection.context.history[-1])
            entry[1] = True
            self._connection.context.history[-1] = tuple(entry)
            return results

    def _execute_statements(
        self,
        operation: str,
        params: Optional[Iterable] = None,
        visibility_filters: Optional[Dict[str, Any]] = None,
    ):
        """
        Executes one or more SQL statements, properly handling comments, cleaning, and splitting.

        Parameters:
            operation: str
                SQL operation(s) to be executed.
            params: Iterable, optional
                Parameters for the SQL operation(s), defaults to None.

        Returns:
            Results of the query execution, if any.
        """
        self._statistics.start_time = time.time_ns()

        if hasattr(operation, "decode"):
            operation = operation.decode()

        operation = sql.remove_comments(operation)
        operation = sql.clean_statement(operation)
        statements = sql.split_sql_statements(operation)

        if len(statements) == 0:
            raise MissingSqlStatement("No statement found")

        if len(statements) > 1 and params is not None and not isinstance(params, dict) and params:
            raise UnsupportedSyntaxError(
                "Batched queries cannot be parameterized with parameter lists, use named parameters."
            )

        results = None
        for index, statement in enumerate(statements):
            results = self._inner_execute(statement, params, visibility_filters)
            if index < len(statements) - 1:
                for _ in results:
                    pass

        # we only return the last result set
        return results

    def execute(
        self,
        operation: str,
        params: Optional[Iterable] = None,
        visibility_filters: Optional[Dict[str, Any]] = None,
    ):
        """
        Executes the provided SQL operation, converting results to internal DataFrame format.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable, optional
                Parameters for the SQL operation, defaults to None.
        """
        self._ensure_open()
        results = self._execute_statements(operation, params, visibility_filters)
        if results is not None:
            result_data, self._result_type = results
            if self._result_type == ResultType.NON_TABULAR:
                import orso

                meta_dataframe = orso.DataFrame(
                    rows=[(result_data.record_count,)],  # type: ignore
                    schema=RelationSchema(
                        name="table",
                        columns=[FlatColumn(name="rows_affected", type=OrsoTypes.INTEGER)],
                    ),
                )  # type: ignore
                self._rows = meta_dataframe._rows
                self._schema = meta_dataframe._schema

                self._rowcount = result_data.record_count  # type: ignore
                self._query_status = result_data.status  # type: ignore
            elif self._result_type == ResultType.TABULAR:
                self._rows, self._schema = converters.from_arrow(result_data)
                self._cursor = iter(self._rows)
                self._query_status = QueryStatus.SQL_SUCCESS
            else:  # pragma: no cover
                self._query_status = QueryStatus.SQL_FAILURE
            self._description = self._schema_to_description(self._schema)
        else:
            self._description = None
        self._executed = True

    @property
    def result_type(self) -> ResultType:
        return self._result_type

    @property
    def query_status(self) -> QueryStatus:
        return self._query_status

    @property
    def rowcount(self) -> int:
        if self._result_type == ResultType.TABULAR:
            return super().rowcount
        if self._result_type == ResultType.NON_TABULAR:
            return self._rowcount
        raise InvalidCursorStateError("Cursor not in valid state to return a row count.")

    @property
    def description(self) -> Optional[Tuple[Tuple[Any, ...], ...]]:
        """DBAPI-compatible column description metadata."""
        return self._description

    def execute_to_arrow(
        self,
        operation: str,
        params: Optional[Iterable] = None,
        limit: Optional[int] = None,
        visibility_filters: Optional[Dict[str, Any]] = None,
    ) -> pyarrow.Table:
        """
        Executes the SQL operation, bypassing conversion to Orso and returning directly in Arrow format.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable, optional
                Parameters for the SQL operation, defaults to None.
            limit: int, optional
                Limit on the number of records to return, defaults to all records.

        Returns:
            The query results in Arrow table format.
        """
        self._ensure_open()
        results = self._execute_statements(operation, params, visibility_filters)
        if results is not None:
            result_data, self._result_type = results

            if self._result_type == ResultType.NON_TABULAR:
                import orso

                meta_dataframe = orso.DataFrame(
                    rows=[(result_data.record_count,)],  # type: ignore
                    schema=RelationSchema(
                        name="table",
                        columns=[FlatColumn(name="rows_affected", type=OrsoTypes.INTEGER)],
                    ),
                )  # type: ignore
                self._executed = True
                return meta_dataframe.arrow()

            if limit is not None:
                result_data = utils.arrow.limit_records(result_data, limit)  # type: ignore

        if isinstance(result_data, pyarrow.Table):
            self._executed = True
            return result_data
        try:
            # arrow allows duplicate column names, but not when concatting
            from itertools import chain

            first_table = next(result_data, None)
            if first_table is not None:
                column_names = first_table.column_names
                if len(column_names) != len(set(column_names)):
                    temporary_names = [f"col_{i}" for i in range(len(column_names))]
                    first_table = first_table.rename_columns(temporary_names)
                    return_table = pyarrow.concat_tables(
                        chain(
                            [first_table], (t.rename_columns(temporary_names) for t in result_data)
                        ),
                        promote_options="permissive",
                    )
                    return return_table.rename_columns(column_names)
            table = pyarrow.concat_tables(
                chain([first_table], result_data), promote_options="permissive"
            )
            self._executed = True
            return table
        except (
            pyarrow.ArrowInvalid,
            pyarrow.ArrowTypeError,
        ) as err:  # pragma: no cover
            # DEBUG: print(err)
            if "struct" in str(err):
                raise InconsistentSchemaError(
                    "Unable to resolve different schemas, most likely related to a STRUCT column."
                ) from err

            from opteryx.exceptions import DataError

            raise DataError(f"Unable to build result dataset ({err})") from err

    @property
    def stats(self) -> Dict[str, Any]:
        """
        Gets the execution statistics.

        Returns:
            Dictionary containing query execution statistics.
        """
        if self._statistics.end_time == 0:  # pragma: no cover
            self._statistics.end_time = time.time_ns()
        return self._statistics.as_dict()

    def execute_to_arrow_batches(
        self,
        operation: str,
        params: Optional[Iterable] = None,
        batch_size: int = 1024,
        limit: Optional[int] = None,
        visibility_filters: Optional[Dict[str, Any]] = None,
    ):
        """
        Execute a SQL operation and stream pyarrow.RecordBatch objects.

        This function mirrors execute_to_arrow but yields RecordBatches in
        a streaming fashion and does not materialize the entire dataset in memory.

        Parameters:
            operation: SQL operation to be executed.
            params: Optional parameters for parameterized queries.
            batch_size: Number of rows per arrow record batch.
            limit: Optional limit on the number of rows to return.
        """
        self._ensure_open()
        results = self._execute_statements(operation, params, visibility_filters)
        if results is None:
            return
        result_data, self._result_type = results

        # Handle non-tabular results (e.g., SET operations)
        if self._result_type == ResultType.NON_TABULAR:
            import orso

            meta_dataframe = orso.DataFrame(
                rows=[(result_data.record_count,)],  # type: ignore
                schema=RelationSchema(
                    name="table",
                    columns=[FlatColumn(name="rows_affected", type=OrsoTypes.INTEGER)],
                ),
            )  # type: ignore
            table = meta_dataframe.arrow()
            self._executed = True
            # update description and state
            self._schema = meta_dataframe._schema
            self._description = self._schema_to_description(self._schema)
            self._query_status = QueryStatus.SQL_SUCCESS
            for batch in table.to_batches(max_chunksize=batch_size):
                yield batch
            return

        # If we have a single pyarrow.Table, iterate over its batches
        if isinstance(result_data, pyarrow.Table):
            table = result_data
            if limit is not None:
                # Limit by slicing rows first, then yield batches
                table = table.slice(offset=0, length=limit)
            self._executed = True
            # set schema and description from this table so users can interrogate cursor
            schema = table.schema
            self._schema = RelationSchema(
                name="table",
                columns=[FlatColumn.from_arrow(field) for field in schema],
            )
            self._description = self._schema_to_description(self._schema)
            self._query_status = QueryStatus.SQL_SUCCESS
            for batch in table.to_batches(max_chunksize=batch_size):
                yield batch
            return

        # For a generator/iterator of pyarrow.Tables, optionally apply a limit and then
        # yield batches from each morsel. We MUST NOT materialize the whole dataset.
        morsels = result_data
        if limit is not None:
            morsels = utils.arrow.limit_records(morsels, limit)

        last_morsel = None
        # buffer of RecordBatches that are not yet large enough to emit
        buffer_batches = []
        buffered_rows = 0

        def _consume_buffered_rows(target_rows: int):
            """
            Consume `target_rows` rows from the buffer_batches and return a pyarrow.RecordBatch.
            This mutates buffer_batches and decreases buffered_rows accordingly.
            """
            nonlocal buffer_batches
            nonlocal buffered_rows
            rows_to_consume = target_rows
            slices = []
            # We will take slices from the start of buffer_batches until we have taken target_rows
            while rows_to_consume > 0 and buffer_batches:
                b = buffer_batches[0]
                if b.num_rows <= rows_to_consume:
                    slices.append(b)
                    rows_to_consume -= b.num_rows
                    buffer_batches.pop(0)
                else:
                    # take required rows from start and keep the remainder
                    slices.append(b.slice(offset=0, length=rows_to_consume))
                    buffer_batches[0] = b.slice(
                        offset=rows_to_consume, length=b.num_rows - rows_to_consume
                    )
                    rows_to_consume = 0

            if not slices:
                return None

            # Convert RecordBatch slices into a combined Table then a single RecordBatch
            # Handle duplicate column names similar to execute_to_arrow
            column_names = slices[0].schema.names
            if len(column_names) != len(set(column_names)):
                temporary_names = [f"col_{i}" for i in range(len(column_names))]
                from itertools import chain

                first_table = slices[0].to_table().rename_columns(temporary_names)
                combined = pyarrow.concat_tables(
                    chain(
                        [first_table],
                        (b.to_table().rename_columns(temporary_names) for b in slices[1:]),
                    ),
                    promote_options="permissive",
                )
                combined = combined.rename_columns(column_names)
                combined = combined.combine_chunks()
            else:
                combined = pyarrow.Table.from_batches(slices).combine_chunks()
            batches = combined.to_batches(max_chunksize=target_rows)
            batch = batches[0] if batches else None
            # update buffered_rows
            buffered_rows = sum(b.num_rows for b in buffer_batches)
            return batch

        for morsel in morsels:
            last_morsel = morsel
            if morsel is None:
                continue
            # set schema and description on the first morsel so users can inspect cursor
            if not getattr(self._schema, "columns", None):
                self._schema = RelationSchema(
                    name="table",
                    columns=[FlatColumn.from_arrow(field) for field in morsel.schema],
                )
                self._description = self._schema_to_description(self._schema)
                self._query_status = QueryStatus.SQL_SUCCESS

            # iterate incoming morsel record batches and accumulate
            for morsel_batch in morsel.to_batches(max_chunksize=batch_size):
                buffer_batches.append(morsel_batch)
                buffered_rows += morsel_batch.num_rows
                while buffered_rows >= batch_size:
                    batch = _consume_buffered_rows(batch_size)
                    if batch is not None:
                        self._executed = True
                        yield batch
                    else:
                        break
            # proceed to next morsel

        # End of result stream: if there's anything left buffered, emit a final batch
        if buffered_rows > 0:
            # take everything that remains
            combined = pyarrow.Table.from_batches(buffer_batches).combine_chunks()
            # last chunk - convert to record batches and yield each (should be <= batch_size)
            for batch in combined.to_batches(max_chunksize=batch_size):
                self._executed = True
                yield batch
        else:
            # if nothing was yielded and we got at least a last_morsel, ensure cursor description & state
            if last_morsel is not None and not self._executed:
                self._schema = RelationSchema(
                    name="table",
                    columns=[FlatColumn.from_arrow(field) for field in last_morsel.schema],
                )
                self._description = self._schema_to_description(self._schema)
                self._query_status = QueryStatus.SQL_SUCCESS

        # Mark executed if we emitted at least one morsel or had a last morsel
        if last_morsel is not None:
            self._executed = True

    @property
    def messages(self) -> List[str]:
        """
        Gets the list of run-time warnings.

        Returns:
            List of warnings generated during query execution.
        """
        return self._statistics.messages

    def close(self):
        """
        Closes the cursor, releasing any resources.
        """
        if self._closed:
            return
        self._cursor = iter(())
        self._description = None
        connection = self._connection
        self._connection = None
        if connection is not None:
            connection._unregister_cursor(self)
            if self._owns_connection:
                connection.close()
        self._closed = True

    def __enter__(self):
        """Support context manager usage for cursors."""
        self._ensure_open()
        return self

    def __exit__(self, exc_type, exc, traceback):
        self.close()
        return False

    def _close_from_connection(self):
        """Called by the Connection when it is closing."""
        self._cursor = iter(())
        self._description = None
        self._connection = None
        self._closed = True

    def __repr__(self):  # pragma: no cover
        """
        Override the Orso repr

        In notebooks we should return a table
        """
        in_a_notebook = False
        try:
            from IPython import get_ipython

            in_a_notebook = get_ipython() is not None
            if not in_a_notebook:
                return f"<opteryx.Cursor {self._state}>"
        except Exception:  # nosec
            pass
        return str(self)

    def __bool__(self):
        """
        Truthy if executed, Falsy if not executed or error
        """
        return self._executed and not self._closed

    def _ensure_open(self):
        if self._closed or self._connection is None:
            raise InvalidCursorStateError("Cursor is closed.")

    @staticmethod
    def _schema_to_description(schema: Optional[RelationSchema]):
        if schema is None or not schema.columns:
            return None
        description: List[Tuple[Any, ...]] = []
        for column in schema.columns:
            description.append(
                (
                    column.name,
                    column.type,
                    None,
                    None,
                    None,
                    None,
                    getattr(column, "nullable", None),
                )
            )
        return tuple(description)



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import numpy
cimport numpy

from libc.stdint cimport int64_t, int32_t
from libc.stddef cimport size_t
from libcpp.vector cimport vector

cdef extern from "intbuffer.h" namespace "" nogil:
    cdef cppclass CIntBuffer:
        CIntBuffer(size_t size_hint) except +
        void append(int64_t value)
        void extend(const vector[int64_t]& values)
        void extend(const int64_t* values, size_t count)
        void reserve(size_t additional_capacity)
        void resize(size_t new_size)
        const int64_t* data() const
        int64_t* mutable_data()
        size_t size() const
        void append_repeated(int64_t value, size_t count)

    cdef cppclass CInt32Buffer:
        CInt32Buffer(size_t size_hint) except +
        void append(int32_t value)
        void extend(const vector[int32_t]& values)
        void extend(const int32_t* values, size_t count)
        void reserve(size_t additional_capacity)
        void resize(size_t new_size)
        const int32_t* data() const
        int32_t* mutable_data()
        size_t size() const


cdef class IntBuffer:

    cdef CIntBuffer* c_buffer
    cdef Py_ssize_t _shape[1]
    cdef Py_ssize_t _strides[1]

    cpdef void append(self, int64_t value)
    cpdef void extend(self, iterable)
    cpdef numpy.ndarray[int64_t, ndim=1] to_numpy(self)
    cpdef const int64_t[::1] get_buffer(self)
    cpdef size_t size(self)
    cpdef void extend_numpy(self, numpy.ndarray[int64_t, ndim=1] arr)
    cpdef void reserve(self, size_t capacity)
    cpdef void append_batch(self, int64_t[::1] values)

cdef class Int32Buffer:

    cdef CInt32Buffer* c_buffer
    cdef Py_ssize_t _shape[1]
    cdef Py_ssize_t _strides[1]

    cpdef void append(self, int32_t value)
    cpdef void extend(self, iterable)
    cpdef numpy.ndarray[int32_t, ndim=1] to_numpy(self)
    cpdef size_t size(self)
    cpdef void reserve(self, size_t capacity)
    cpdef void extend_numpy(self, numpy.ndarray[int32_t, ndim=1] arr)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

from opteryx.managers.kvstores.base_kv_store import BaseKeyValueStore
from opteryx.managers.kvstores.factory import create_kv_store
from opteryx.managers.kvstores.file_kv_store import FileKeyValueStore
from opteryx.managers.kvstores.gcs_kv_store import GCSKeyValueStore
from opteryx.managers.kvstores.memcached import MemcachedCache
from opteryx.managers.kvstores.null_cache import NullCache
from opteryx.managers.kvstores.redis import RedisCache
from opteryx.managers.kvstores.s3_kv_store import S3KeyValueStore
from opteryx.managers.kvstores.valkey import ValkeyCache

__all__ = [
    "BaseKeyValueStore",
    "FileKeyValueStore",
    "S3KeyValueStore",
    "GCSKeyValueStore",
    "RedisCache",
    "MemcachedCache",
    "NullCache",
    "ValkeyCache",
    "create_kv_store",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

import datetime
from typing import Callable
from typing import List
from typing import Optional

from opteryx.managers.schemes import BasePartitionScheme
from opteryx.utils.file_decoders import KNOWN_EXTENSIONS
from opteryx.utils.file_decoders import ExtentionType


class DefaultPartitionScheme(BasePartitionScheme):
    def get_blobs_in_partition(
        self,
        *,
        blob_list_getter: Callable,
        prefix: str,
        start_date: Optional[datetime.datetime],
        end_date: Optional[datetime.datetime],
        **kwargs,
    ) -> List[str]:
        data_exts = tuple(
            ext for ext, (_, ext_type) in KNOWN_EXTENSIONS.items() if ext_type == ExtentionType.DATA
        )
        list_of_blobs = blob_list_getter(prefix=prefix)
        return [name for name in list_of_blobs if name.endswith(data_exts)]



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

"""
Ultra-fast disk reader module
"""

import errno

from cpython.buffer cimport PyBuffer_FillInfo
from cpython.mem cimport PyMem_Free
from cpython.mem cimport PyMem_Malloc
from cpython.unicode cimport PyUnicode_FromString
from libc.stddef cimport size_t

cdef extern from "disk_io.h":
    int read_all_pread(const char* path, unsigned char* dst, size_t* out_len,
                       bint sequential, bint willneed, bint drop_after)
    int read_all_mmap(const char* path, unsigned char** dst, size_t* out_len)
    int unmap_memory_c(unsigned char* addr, size_t size)

cdef extern from "directories.h":
    ctypedef struct file_info_t:
        char* name
        int is_directory
        int is_regular_file
        long long size
        long long mtime

    int list_directory_c "list_directory"(const char* path, file_info_t** files, size_t* count)
    void free_file_list(file_info_t* files, size_t count)
    int list_matching_files_c "list_matching_files"(const char* base_path, const char** extensions,
                                                    size_t ext_count, char*** files, size_t* count) nogil
    void free_file_names(char** files, size_t count)

cdef class MappedMemory:
    cdef unsigned char* data
    cdef size_t size
    cdef bint owned

    def __getbuffer__(self, Py_buffer* buffer, int flags):
        PyBuffer_FillInfo(buffer, self, self.data, self.size, 1, flags)

    def __len__(self):
        return self.size


def read_file(str path, bint sequential=True, bint willneed=True, bint drop_after=False):
    """
    Read an entire file into memory with optimized I/O.
    """
    import os

    if not os.path.exists(path):
        raise FileNotFoundError(f"File not found: {path}")

    cdef size_t size = os.path.getsize(path)
    cdef size_t out_len = 0

    path_b = path.encode("utf-8")
    cdef const char* c_path = path_b

    # Allocate buffer - use bytearray for mutable buffer
    buf = bytearray(size)
    cdef unsigned char[::1] buf_view = buf
    cdef unsigned char* dst = &buf_view[0]

    cdef int rc = read_all_pread(c_path, dst, &out_len, sequential, willneed, drop_after)

    if rc != 0:
        raise OSError(-rc, f"Failed to read file: {path}")

    return memoryview(buf)[:out_len]


def list_directory(str path):
    """Return directory entries using the native file system scanner."""

    path_b = path.encode("utf-8")
    cdef const char* c_path = path_b
    cdef file_info_t* files = NULL
    cdef size_t count = 0

    cdef int rc = list_directory_c(c_path, &files, &count)
    if rc != 0:
        err = -rc
        if err == errno.ENOENT:
            raise FileNotFoundError(path)
        raise OSError(err, f"Failed to list directory: {path}")

    entries = []
    cdef file_info_t entry
    cdef char* name_ptr
    cdef size_t idx
    try:
        for idx in range(count):
            entry = files[idx]
            name_ptr = entry.name
            if name_ptr == NULL:
                continue

            py_name = PyUnicode_FromString(name_ptr)
            entries.append(
                (
                    py_name,
                    bool(entry.is_directory),
                    bool(entry.is_regular_file),
                    entry.size,
                    entry.mtime,
                )
            )
    finally:
        if files != NULL:
            free_file_list(files, count)

    return entries


def list_files(str path, extensions):
    """Return a list of files under ``path`` matching provided extensions."""

    if extensions is None:
        raise ValueError("extensions must be provided")

    ext_seq = tuple(extensions)
    ext_count = len(ext_seq)

    cdef size_t c_ext_count = ext_count
    cdef const char** ext_array = NULL
    cdef char** file_array = NULL
    cdef size_t file_count = 0

    ext_bytes = [ext.encode("utf-8") if isinstance(ext, str) else ext for ext in ext_seq]

    if c_ext_count > 0:
        ext_array = <const char**>PyMem_Malloc(c_ext_count * sizeof(const char*))
        if ext_array == NULL:
            raise MemoryError("Unable to allocate extension array")
        for idx in range(c_ext_count):
            ext_array[idx] = <const char*>ext_bytes[idx]

    path_b = path.encode("utf-8")
    cdef const char* c_path = path_b

    cdef int rc
    with nogil:
        rc = list_matching_files_c(c_path, ext_array, c_ext_count, &file_array, &file_count)

    if ext_array != NULL:
        PyMem_Free(ext_array)

    if rc != 0:
        if file_array != NULL:
            free_file_names(file_array, file_count)
        err = -rc
        if err == errno.ENOENT:
            raise FileNotFoundError(path)
        raise OSError(err, f"Failed to list files: {path}")

    results = []
    try:
        for idx in range(file_count):
            py_path = PyUnicode_FromString(file_array[idx])
            if py_path is None:
                raise MemoryError("Unable to decode file path")
            results.append(py_path)
    finally:
        if file_array != NULL:
            free_file_names(file_array, file_count)

    return results


def read_file_mmap(str path):
    """
    Read file using memory mapping - returns an object that provides memoryview interface
    but MUST be manually closed to avoid resource leaks.
    """
    import os

    if not os.path.exists(path):
        raise FileNotFoundError(f"File not found: {path}")

    path_b = path.encode("utf-8")
    cdef const char* c_path = path_b
    cdef unsigned char* mapped_data = NULL
    cdef size_t size = 0

    cdef int rc = read_all_mmap(c_path, &mapped_data, &size)

    if rc != 0:
        raise OSError(-rc, f"Failed to mmap file: {path}")

    # Create wrapper that knows how to clean up
    cdef MappedMemory wrapper = MappedMemory.__new__(MappedMemory)
    wrapper.data = mapped_data
    wrapper.size = size
    wrapper.owned = False  # This is mmap'd memory, not malloc'd

    return wrapper


def read_file_to_bytes(str path, bint sequential=True, bint willneed=True, bint drop_after=False):
    """
    Read an entire file into memory as bytes.
    """
    mv = read_file(path, sequential, willneed, drop_after)
    return bytes(mv)


def unmap_memory(mem_obj):
    """
    Explicitly unmap memory from read_file_mmap.
    MUST be called when done with the data to avoid resource leaks.
    """
    cdef int rc
    if hasattr(mem_obj, 'data') and mem_obj.data is not None:
        # Import the unmap function from your C code
        rc = unmap_memory_c(mem_obj.data, mem_obj.size)
        mem_obj.data = None
        return rc == 0
    return True



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.


import asyncio
from functools import wraps

from opteryx.config import MAX_CACHE_EVICTIONS_PER_QUERY
from opteryx.config import MAX_CACHEABLE_ITEM_SIZE
from opteryx.third_party.cyan4973.xxhash import hash_bytes

__all__ = ("Cacheable", "async_read_thru_cache", "read_thru_cache")

SOURCE_NOT_FOUND = 0
SOURCE_BUFFER_POOL = 1
SOURCE_REMOTE_CACHE = 2
SOURCE_ORIGIN = 3


class Cacheable:
    """
    This class is just a marker - it is empty.

    Caching is added in the binding phase.
    """

    def __init__(self, *args, **kwargs):
        pass

    def read_blob(self, *, blob_name, **kwargs):
        pass

    def purge_blob(self, blob_name: str):
        """
        Purge the blob from the cache.

        This is used to remove blobs that are invalid or no longer needed.
        """
        from opteryx import get_cache_manager
        from opteryx.shared import BufferPool

        key = hex(hash_bytes(blob_name.encode())).encode()

        try:
            # Purge from the local buffer pool
            buffer_pool = BufferPool()
            buffer_pool.delete(key)

            # Purge from the remote cache
            cache_manager = get_cache_manager()
            remote_cache = cache_manager.cache_backend
            if remote_cache:
                remote_cache.delete(key)
        finally:
            # best endeavours to remove the blob from the cache
            return None


def read_thru_cache(func):
    """
    Decorator to implement a read-thru cache.

    It intercepts requests to read blobs and first looks them up in the in-memory
    cache (BufferPool) and optionally in a secondary cache (like MemcacheD or Redis).
    """

    # Capture the max_evictions value at decoration time
    from opteryx import get_cache_manager
    from opteryx.managers.cache import NullCache
    from opteryx.shared import BufferPool

    cache_manager = get_cache_manager()
    max_evictions = MAX_CACHE_EVICTIONS_PER_QUERY
    remote_cache = cache_manager.cache_backend
    if not remote_cache:
        # rather than make decisions - just use a dummy
        remote_cache = NullCache()

    buffer_pool = BufferPool()

    my_keys = set()

    @wraps(func)
    def wrapper(blob_name, statistics, **kwargs):
        nonlocal max_evictions

        key = hex(hash_bytes(blob_name.encode())).encode()
        my_keys.add(key)

        # try the buffer pool first
        result = buffer_pool.get(key)
        if result is not None:
            statistics.bufferpool_hits += 1
            return result

        # try the remote cache next
        result = remote_cache.get(key)
        if result is not None:
            statistics.remote_cache_hits += 1
            return result

        # Key is not in cache, execute the function and store the result in cache
        result = func(blob_name=blob_name, **kwargs)

        # Write the result to caches
        if max_evictions:
            # we set a per-query eviction limit
            if len(result) < MAX_CACHEABLE_ITEM_SIZE:
                evicted = buffer_pool.set(key, result)
                remote_cache.set(key, result)
                if evicted:
                    # if we're evicting items we're putting into the cache
                    if evicted in my_keys:
                        max_evictions = 0
                    else:
                        max_evictions -= 1
                    statistics.cache_evictions += 1
            else:
                statistics.cache_oversize += 1

        statistics.cache_misses += 1

        return result

    return wrapper


def async_read_thru_cache(func):
    """
    This is added to the reader by the binder.

    We check the faster buffer pool (local memory), then the cache (usually
    remote cache like memcached) before fetching from source.

    The async readers don't return the bytes, instead they return the
    read buffer reference for the bytes, which means we need to populate
    the read buffer and return the ref for these items.
    """
    # Capture the evictions_remaining value at decoration time
    from opteryx import get_cache_manager
    from opteryx import system_statistics
    from opteryx.managers.cache import NullCache
    from opteryx.shared import BufferPool
    from opteryx.shared import MemoryPool

    cache_manager = get_cache_manager()
    evictions_remaining = MAX_CACHE_EVICTIONS_PER_QUERY
    remote_cache = cache_manager.cache_backend
    if not remote_cache:
        # rather than make decisions - just use a dummy
        remote_cache = NullCache()

    buffer_pool = BufferPool()

    my_keys = set()

    @wraps(func)
    async def wrapper(blob_name: str, statistics, pool: MemoryPool, **kwargs):
        try:
            nonlocal evictions_remaining

            source = SOURCE_NOT_FOUND
            key = hex(hash_bytes(blob_name.encode())).encode()
            read_buffer_ref = None
            payload = None
            my_keys.add(key)

            # try the buffer pool first
            payload = buffer_pool.get(key, zero_copy=True, latch=True)
            if payload is not None:
                source = SOURCE_BUFFER_POOL
                remote_cache.touch(key)  # help the remote cache track LRU
                statistics.bufferpool_hits += 1
                read_buffer_ref = await pool.commit(payload)  # type: ignore
                while read_buffer_ref == -1:  # pragma: no cover
                    await asyncio.sleep(0.1)
                    statistics.stalls_io_waiting_on_engine += 1
                    read_buffer_ref = await pool.commit(payload)  # type: ignore
                    statistics.bytes_read += len(payload)
                    system_statistics.cpu_wait_seconds += 0.1
                buffer_pool.unlatch(key)
                return read_buffer_ref

            # try the remote cache next
            payload = remote_cache.get(key)
            if payload is not None:
                source = SOURCE_REMOTE_CACHE
                statistics.remote_cache_hits += 1
                system_statistics.remote_cache_reads += 1
                read_buffer_ref = await pool.commit(payload)  # type: ignore
                while read_buffer_ref == -1:  # pragma: no cover
                    await asyncio.sleep(0.1)
                    statistics.stalls_io_waiting_on_engine += 1
                    read_buffer_ref = await pool.commit(payload)  # type: ignore
                    statistics.bytes_read += len(payload)
                    system_statistics.cpu_wait_seconds += 0.1
                return read_buffer_ref

            try:
                read_buffer_ref = await func(
                    blob_name=blob_name, statistics=statistics, pool=pool, **kwargs
                )
                source = SOURCE_ORIGIN
                statistics.cache_misses += 1
                system_statistics.origin_reads += 1
                return read_buffer_ref
            except Exception as e:  # pragma: no cover
                print(f"Error in {func.__name__}: {e}")
                raise  # Optionally re-raise the error after logging it

        finally:
            if payload is None and read_buffer_ref is not None:
                # we set a per-query eviction limit
                payload = await pool.read(read_buffer_ref)  # type: ignore

            # If we found the file, see if we need to write it to the caches
            if (
                not source in (SOURCE_NOT_FOUND, SOURCE_BUFFER_POOL)
                and evictions_remaining > 0
                and len(payload) < buffer_pool.size // 10
            ):
                # if we didn't get it from the buffer pool (origin or remote cache) we add it
                evicted = buffer_pool.set(key, payload)
                if evicted:  # pragma: no cover
                    # if we're evicting items we just put in the cache, stop
                    if evicted in my_keys:
                        evictions_remaining = -1
                    else:
                        evictions_remaining -= 1
                    statistics.cache_evictions += 1

            if payload is None or isinstance(remote_cache, NullCache):
                # If we didn't find the payload, we don't write it to the cache
                pass
            elif source == SOURCE_ORIGIN and len(payload) < MAX_CACHEABLE_ITEM_SIZE:
                # If we read from the source, it's not in the remote cache
                remote_cache.set(key, payload)
                system_statistics.remote_cache_commits += 1
            elif len(payload) >= MAX_CACHEABLE_ITEM_SIZE:
                statistics.cache_oversize += 1

    return wrapper



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import numpy
cimport numpy
numpy.import_array()

from opteryx.third_party.abseil.containers cimport FlatHashSet
from opteryx.third_party.fastfloat.fast_float cimport c_parse_fast_float as parse_fast_float

from libc.stdint cimport int32_t, int64_t, uint64_t, uint8_t, uintptr_t
from cpython.unicode cimport PyUnicode_DecodeUTF8
from cpython.object cimport PyObject_Hash
from cpython.bytes cimport PyBytes_FromStringAndSize

cpdef tuple build_rows_indices_and_column(object column):
    cdef:
        object child_elements = column.values
        list buffers = column.buffers()
        Py_ssize_t row_count = len(column)
        Py_ssize_t total_size
        Py_ssize_t i
        Py_ssize_t j
        Py_ssize_t index_pos
        numpy.ndarray indices_np, flat_data_np
        int64_t[::1] indices
        object[:] flat_data
        Py_ssize_t arr_offset = column.offset
        const int32_t* offsets32 = <const int32_t*><uintptr_t>(buffers[1].address)

        # Child array variables
        Py_ssize_t child_offset = child_elements.offset
        list child_buffers = child_elements.buffers()
        const int32_t* child_offsets32 = <const int32_t*><uintptr_t>(child_buffers[1].address)
        const char* child_data = <const char*><uintptr_t>(child_buffers[2].address)

        const uint8_t* parent_valid = <const uint8_t*><uintptr_t>(buffers[0].address) if buffers[0] else NULL
        const uint8_t* child_valid = <const uint8_t*><uintptr_t>(child_buffers[0].address) if child_buffers[0] else NULL

        Py_ssize_t start
        Py_ssize_t end
        Py_ssize_t str_start
        Py_ssize_t str_end

    if row_count == 0:
        return numpy.empty(0, dtype=numpy.int64), numpy.empty(0, dtype=object)

    total_size = offsets32[arr_offset + row_count] - offsets32[arr_offset]
    if total_size == 0:
        return numpy.empty(0, dtype=numpy.int64), numpy.empty(0, dtype=object)

    indices_np = numpy.empty(total_size, dtype=numpy.int64)
    flat_data_np = numpy.empty(total_size, dtype=object)
    indices = indices_np
    flat_data = flat_data_np

    index_pos = 0
    for i in range(row_count):
        if parent_valid and not (parent_valid[i >> 3] & (1 << (i & 7))):
            continue

        start = offsets32[arr_offset + i]
        end = offsets32[arr_offset + i + 1]

        if start >= end:
            continue

        for j in range(start, end):
            indices[index_pos] = i

            if child_valid and not (child_valid[(child_offset + j) >> 3] & (1 << ((child_offset + j) & 7))):
                flat_data[index_pos] = None
            else:
                str_start = child_offsets32[child_offset + j]
                str_end = child_offsets32[child_offset + j + 1]

                if str_end > str_start:
                    flat_data[index_pos] = PyUnicode_DecodeUTF8(
                        child_data + str_start, str_end - str_start, "replace"
                    )
                else:
                    flat_data[index_pos] = ""
            index_pos += 1

    return indices_np, flat_data_np


cpdef tuple numpy_build_rows_indices_and_column(numpy.ndarray column_data):
    cdef int64_t row_count = column_data.shape[0]
    cdef numpy.int64_t[::1] lengths = numpy.empty(row_count, dtype=numpy.int64)
    cdef numpy.int64_t[::1] offsets = numpy.empty(row_count + 1, dtype=numpy.int64)
    cdef int64_t i
    cdef int64_t total_size = 0
    cdef numpy.dtype element_dtype = column_data[0].dtype

    if not isinstance(column_data[0], numpy.ndarray):
        raise TypeError("UNNEST requires an ARRAY column.")

    # Calculate lengths and total_size
    for i in range(row_count):
        lengths[i] = column_data[i].shape[0]
        total_size += lengths[i]

    # Early exit if total_size is zero
    if total_size == 0:
        return (numpy.array([], dtype=numpy.int64), numpy.array([], dtype=object))

    # Compute offsets for efficient slicing
    offsets[0] = 0
    for i in range(row_count):
        offsets[i + 1] = offsets[i] + lengths[i]
    cdef numpy.int64_t[::1] indices = numpy.empty(total_size, dtype=numpy.int64)
    cdef numpy.ndarray flat_data = numpy.empty(total_size, dtype=element_dtype)

    # Fill indices and flat_data
    for i in range(row_count):
        start = offsets[i]
        end = offsets[i + 1]
        if end > start:
            indices[start:end] = i
            flat_data[start:end] = column_data[i]

    return (indices, flat_data)


cpdef tuple numpy_build_filtered_rows_indices_and_column(numpy.ndarray column_data, set valid_values):
    """
    Build row indices and flattened column data for matching values from a column of array-like elements.

    Parameters:
        column_data: ndarray
            An array of arrays from which to create row indices and flattened data.
        valid_values: set
            A set of values to filter the rows by during the cross join.

    Returns:
        tuple of (ndarray, ndarray)
            Returns a tuple containing an array of indices and an array of flattened data for rows that match the filter.
    """
    cdef int64_t row_count = column_data.shape[0]
    cdef int64_t allocated_size = row_count * 4  # Initial allocation size
    cdef int64_t index = 0
    cdef int64_t i, j, len_i
    cdef object array_i
    cdef numpy.ndarray flat_data
    cdef numpy.int64_t[::1] indices
    cdef numpy.dtype element_dtype = numpy.dtype(object)
    cdef object value

    # Typed sets for different data types
    cdef set valid_values_typed = None

    # Determine the dtype of the elements
    for i in range(row_count):
        array_i = column_data[i]
        if array_i is not None and array_i.size > 0:
            element_dtype = array_i.dtype
            break

    # Initialize indices and flat_data arrays
    indices = numpy.empty(allocated_size, dtype=numpy.int64)
    flat_data = numpy.empty(allocated_size, dtype=element_dtype)

    # Handle set initialization based on element dtype
    if numpy.issubdtype(element_dtype, numpy.integer):
        valid_values_typed = {int(v) for v in valid_values}
    elif numpy.issubdtype(element_dtype, numpy.floating):
        valid_values_typed = {parse_fast_float(v) for v in valid_values}
    elif numpy.issubdtype(element_dtype, numpy.str_):
        valid_values_typed = {unicode(v) for v in valid_values}
    else:
        valid_values_typed = valid_values  # Fallback to generic Python set

    # Main loop
    for i in range(row_count):
        array_i = column_data[i]
        if array_i is None:
            continue
        len_i = array_i.shape[0]
        if len_i == 0:
            continue

        for j in range(len_i):
            value = array_i[j]
            if value in valid_values_typed:
                if index >= allocated_size:
                    # Reallocate arrays
                    allocated_size *= 2
                    indices = numpy.resize(indices, allocated_size)
                    flat_data = numpy.resize(flat_data, allocated_size)
                flat_data[index] = value
                indices[index] = i
                index += 1

    if index == 0:
        return (numpy.array([], dtype=numpy.int64), numpy.array([], dtype=element_dtype))

    # Slice arrays to the actual used size
    indices = indices[:index]
    flat_data = flat_data[:index]

    return (indices, flat_data)


cpdef tuple build_filtered_rows_indices_and_column(object column, set valid_values):
    """
    Arrow-native version of build_filtered_rows_indices_and_column.
    Filters values from a ListArray column based on membership in `valid_values`.
    Returns matching row indices and values (as bytes, not str).
    """
    cdef:
        object child_elements = column.values
        list buffers = column.buffers()
        Py_ssize_t row_count = len(column)
        Py_ssize_t arr_offset = column.offset
        const int32_t* offsets32 = <const int32_t*><uintptr_t>(buffers[1].address)
        Py_ssize_t i = 0
        Py_ssize_t j = 0
        Py_ssize_t k = 0
        Py_ssize_t start
        Py_ssize_t end
        Py_ssize_t str_len
        Py_ssize_t str_end
        Py_ssize_t allocated_size = row_count * 4 if row_count > 0 else 4

        numpy.ndarray indices = numpy.empty(allocated_size, dtype=numpy.int64)
        int64_t[::1] indices_mv = indices
        numpy.ndarray flat_data = numpy.empty(allocated_size, dtype=object)
        object[::1] flat_mv = flat_data

        list child_buffers = child_elements.buffers()
        const int32_t* child_offsets32 = <const int32_t*><uintptr_t>(child_buffers[1].address)
        const char* child_data = <const char*><uintptr_t>(child_buffers[2].address)
        Py_ssize_t child_offset = child_elements.offset
        const uint8_t* parent_bitmap = NULL
        const uint8_t* child_bitmap = NULL

        Py_ssize_t str_start

    if buffers[0]:
        parent_bitmap = <const uint8_t*><uintptr_t>(buffers[0].address)
    if child_buffers[0]:
        child_bitmap = <const uint8_t*><uintptr_t>(child_buffers[0].address)

    # Normalize valid_values to bytes
    cdef set valid_bytes = set()
    for v in valid_values:
        valid_bytes.add(v.encode("utf8") if isinstance(v, str) else v)

    for i in range(row_count):
        if parent_bitmap is not NULL and not (parent_bitmap[i >> 3] & (1 << (i & 7))):
            continue

        start = offsets32[arr_offset + i]
        end = offsets32[arr_offset + i + 1]

        for j in range(start, end):
            if child_bitmap is not NULL and not (child_bitmap[(child_offset + j) >> 3] & (1 << ((child_offset + j) & 7))):
                continue

            str_start = child_offsets32[child_offset + j]
            str_end = child_offsets32[child_offset + j + 1]
            str_len = str_end - str_start

            # Materialize only matched values as bytes
            value_bytes = PyBytes_FromStringAndSize(child_data + str_start, str_len)

            if value_bytes in valid_bytes:
                if k >= allocated_size:
                    allocated_size *= 2
                    new_indices = numpy.empty(allocated_size, dtype=numpy.int64)
                    new_indices[:k] = indices_mv[:k]
                    indices = new_indices
                    indices_mv = indices

                    new_flat = numpy.empty(allocated_size, dtype=object)
                    new_flat[:k] = flat_data[:k]
                    flat_data = new_flat
                    flat_mv = flat_data

                flat_mv[k] = value_bytes
                indices_mv[k] = i
                k += 1

    return indices_mv[:k], flat_mv[:k]


cpdef tuple list_distinct(numpy.ndarray values, int64_t[::1] indices, FlatHashSet seen_hashes=None):
    cdef:
        Py_ssize_t i = 0
        Py_ssize_t j = 0
        Py_ssize_t n = values.shape[0]
        uint64_t hash_value
        object v
        numpy.dtype dtype = values.dtype
        numpy.ndarray new_values = numpy.empty(n, dtype=dtype)
        int64_t[::1] new_indices = numpy.empty(n, dtype=numpy.int64)

    if seen_hashes is None:
        seen_hashes = FlatHashSet()

    for i in range(n):
        v = values[i]
        hash_value = <uint64_t>(PyObject_Hash(v) & 0xFFFFFFFFFFFFFFFF)
        if seen_hashes.insert(hash_value):
            new_values[j] = v
            new_indices[j] = indices[i]
            j += 1

    return new_values[:j], new_indices[:j], seen_hashes



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.



# THIS FILE IS AUTOMATICALLY UPDATED DURING THE BUILD PROCESS
# DO NOT EDIT THIS FILE DIRECTLY

__build__ = 1977
__author__ = "@joocer"
__version__ = "0.26.3"

# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it in the CLI for the same reason



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Outer Join Node

This is a SQL Query Execution Plan Node.

PyArrow has LEFT/RIGHT/FULL OUTER JOIN implementations, but they error when the
relations being joined contain STRUCT or ARRAY columns so we've written our own
OUTER JOIN implementations.
"""

import time
from array import array
from typing import List

import pyarrow

from opteryx import EOS
from opteryx.compiled.joins import build_side_hash_map
from opteryx.compiled.joins import probe_side_hash_map
from opteryx.compiled.joins import right_join
from opteryx.compiled.structures.bloom_filter import create_bloom_filter
from opteryx.compiled.structures.buffers import IntBuffer
from opteryx.compiled.structures.hash_table import HashTable
from opteryx.models import QueryProperties
from opteryx.utils.arrow import align_tables

from . import JoinNode

CHUNK_SIZE: int = 50_000


def left_join(
    left_relation,
    right_relation,
    left_columns: List[str],
    right_columns: List[str],
    filter_index,
    left_hash,
    columns=None,
):
    """
    Perform a LEFT OUTER JOIN using a prebuilt hash map and optional filter.

    Yields:
        pyarrow.Table chunks of the joined result.
    """

    left_indexes = IntBuffer()
    right_indexes = IntBuffer()
    seen_flags = array("b", [0]) * left_relation.num_rows

    if filter_index:
        # We can just dispose of rows from the right relation that don't match
        # our bloom filter
        possibly_matching_rows = filter_index.possibly_contains_many(right_relation, right_columns)
        right_relation = right_relation.filter(possibly_matching_rows)

        # If there's no matching rows in the right relation, we can exit early
        if right_relation.num_rows == 0:
            # Short circuit: no matching right rows at all
            for i in range(0, left_relation.num_rows, CHUNK_SIZE):
                chunk = list(range(i, min(i + CHUNK_SIZE, left_relation.num_rows)))
                yield align_tables(
                    source_table=left_relation,
                    append_table=right_relation.slice(0, 0),
                    source_indices=chunk,
                    append_indices=[None] * len(chunk),
                )
            return

    # Build the hash table of the right relation
    right_hash = probe_side_hash_map(right_relation, right_columns)

    for h, right_rows in right_hash.hash_table.items():
        left_rows = left_hash.get(h)
        if not left_rows:
            continue
        for l in left_rows:
            seen_flags[l] = 1
            left_indexes.extend([l] * len(right_rows))
            right_indexes.extend(right_rows)

    # Yield matching rows
    if left_indexes.size() > 0:
        yield align_tables(
            right_relation,
            left_relation,
            right_indexes.to_numpy(),
            left_indexes.to_numpy(),
        )

    # Emit unmatched left rows using null-filled right columns
    unmatched = [i for i, seen in enumerate(seen_flags) if not seen]

    if unmatched:
        unmatched_left = left_relation.take(pyarrow.array(unmatched))
        # Create a right-side table with zero rows, we do this because
        # we want arrow to do the heavy lifting of adding new columns to
        # the left relation, we do not want to add rows to the left
        # relation - arrow is faster at adding null columns that we can be.
        null_right = pyarrow.table(
            [pyarrow.nulls(0, type=field.type) for field in right_relation.schema],
            schema=right_relation.schema,
        )
        yield pyarrow.concat_tables([unmatched_left, null_right], promote_options="permissive")

    return


def full_join(
    left_relation,
    right_relation,
    left_columns: List[str],
    right_columns: List[str],
    columns=None,
    **kwargs,
):
    hash_table = HashTable()
    non_null_right_values = right_relation.select(right_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*non_null_right_values)):
        hash_table.insert(abs(hash(value_tuple)), i)

    left_indexes = []
    right_indexes = []

    left_values = left_relation.select(left_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*left_values)):
        rows = hash_table.get(abs(hash(value_tuple)))
        if rows:
            right_indexes.extend(rows)
            left_indexes.extend([i] * len(rows))
        else:
            right_indexes.append(None)
            left_indexes.append(i)

    for i in range(right_relation.num_rows):
        if i not in right_indexes:
            right_indexes.append(i)
            left_indexes.append(None)

    for i in range(0, len(left_indexes), CHUNK_SIZE):
        chunk_left_indexes = left_indexes[i : i + CHUNK_SIZE]
        chunk_right_indexes = right_indexes[i : i + CHUNK_SIZE]

        # Align this chunk and add the resulting table to our list
        yield align_tables(right_relation, left_relation, chunk_right_indexes, chunk_left_indexes)


class OuterJoinNode(JoinNode):
    def __init__(self, properties: QueryProperties, **parameters):
        # Ensure `join_type` exists before the base initializer accesses `self.name`
        self.join_type = parameters["type"]
        JoinNode.__init__(self, properties=properties, **parameters)
        self.on = parameters.get("on")
        self.using = parameters.get("using")

        self.left_columns = parameters.get("left_columns")
        self.left_readers = parameters.get("left_readers")

        self.right_columns = parameters.get("right_columns")
        self.right_readers = parameters.get("right_readers")

        self.columns = parameters.get("columns")

        self.left_buffer = []
        self.left_buffer_columns = None
        self.right_buffer = []
        self.left_relation = None
        self.empty_right_relation = None
        self.left_hash = None
        self.left_seen_rows = set()

        self.filter_index = None

    @property
    def name(self):  # pragma: no cover
        return self.join_type.replace(" ", "_")

    @property
    def config(self) -> str:  # pragma: no cover
        from opteryx.managers.expression import format_expression

        if self.on:
            return f"{self.join_type.upper()} JOIN ({format_expression(self.on, True)})"
        if self.using:
            return f"{self.join_type.upper()} JOIN (USING {','.join(map(format_expression, self.using))})"
        return f"{self.join_type.upper()}"

    def execute(self, morsel: pyarrow.Table, join_leg: str) -> pyarrow.Table:
        morsel = self.ensure_arrow_table(morsel)

        if join_leg == "left":
            if morsel == EOS:
                self.left_relation = pyarrow.concat_tables(self.left_buffer, promote_options="none")
                self.left_buffer.clear()
                if self.join_type == "left outer":
                    start = time.monotonic_ns()
                    self.left_hash = build_side_hash_map(self.left_relation, self.left_columns)
                    self.statistics.time_build_hash_map += time.monotonic_ns() - start

                    if self.left_relation.num_rows < 16_000_001:
                        start = time.monotonic_ns()
                        self.filter_index = create_bloom_filter(
                            self.left_relation, self.left_columns
                        )
                        self.statistics.time_build_bloom_filter += time.monotonic_ns() - start
                        self.statistics.feature_bloom_filter += 1
            else:
                if self.left_buffer_columns is None:
                    self.left_buffer_columns = morsel.schema.names
                else:
                    morsel = morsel.select(self.left_buffer_columns)
                self.left_buffer.append(morsel)
            yield None
            return

        if join_leg == "right":
            if morsel == EOS:
                right_relation = pyarrow.concat_tables(self.right_buffer, promote_options="none")
                self.right_buffer.clear()

                join_provider = providers.get(self.join_type)

                for result_table in join_provider(
                    left_relation=self.left_relation,
                    right_relation=right_relation,
                    left_columns=self.left_columns,
                    right_columns=self.right_columns,
                    left_hash=self.left_hash,
                    filter_index=self.filter_index,
                    columns=self.columns,
                ):
                    # Project down to only the needed columns if specified
                    if self.columns is not None:
                        candidates = [c.schema_column.identity for c in self.columns]
                        keep_columns = [c for c in candidates if c in result_table.schema.names]
                        result_table = result_table.select(keep_columns)
                    yield result_table
                yield EOS

            else:
                self.right_buffer.append(morsel)
                yield None


providers = {"left outer": left_join, "full outer": full_join, "right outer": right_join}



"""
Compiled I/O operations for high-performance file access.
"""

try:
    from .disk_reader import read_file
    from .disk_reader import read_file_to_bytes

    __all__ = ["read_file", "read_file_to_bytes"]
except ImportError:
    # Module not yet compiled
    __all__ = []



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

"""
This is not a general perpose Bloom Filter, if used outside Opteryx it may not
perform entirely as expected as it is optimized for a specific configuration
and constraints.

We have four size options, all using 2 hashes:
    - a 8k slot bit array for up to 1000 items (about 4.9% FPR)
    - a 512k slot bit array for up to 60k items (about 4.2% FPR)
    - a 8m slot bit array for up to 1m items (about 4.5% FPR)
    - a 128m slot bit array for up to 16m items (about 4.7% FPR)

We perform one hash and then use a calculation based on the golden ratio to
determine the second position. This is cheaper than performing two hashes whilst
still providing a good enough split of the two hashes.

The primary use for this structure is to prefilter JOINs, it is many times faster
(about 20x from initial benchmarking) to test for containment in the bloom filter
that to look up the item in the hash table.

Building the filter is fast - for tables up to 1 million records we create the filter
(1m records is roughly a 0.005s build). If the filter isn't effective (less that 5%
eliminations) we discard it which has meant some waste work.

The 16m set is the limit at the moment, it takes about 0.08 seconds to build which
is the limit of what we think we should speculatively build.
"""

from libc.stdlib cimport calloc, free
from libc.stdint cimport uint64_t, uint32_t

from opteryx.compiled.table_ops.hash_ops cimport compute_row_hashes
from opteryx.compiled.table_ops.null_avoidant_ops cimport non_null_row_indices

import numpy
cimport numpy

cdef extern from "<stdint.h>":
    ctypedef unsigned long uintptr_t

# Define sizes for the Bloom filters - now in 64-bit chunks
cdef uint32_t BIT64_ARRAY_SIZE_TINY = 128  # 128 * 64 = 8,192 bits
cdef uint32_t BIT64_ARRAY_SIZE_SMALL = 8 * 1024  # 8K * 64 = 524,288 bits
cdef uint32_t BIT64_ARRAY_SIZE_LARGE = 128 * 1024  # 128K * 64 = 8,388,608 bits
cdef uint32_t BIT64_ARRAY_SIZE_HUGE = 2 * 1024 * 1024  # 2M * 64 = 134,217,728 bits

# Golden ratio constant for second hash
cdef uint64_t GOLDEN_RATIO = 0x9E3779B97F4A7C15ULL

cdef class BloomFilter:

    def __cinit__(self, uint32_t expected_records=50000):
        """Initialize Bloom Filter based on expected number of records."""
        if expected_records <= 1_000:
            self.bit64_array_size = BIT64_ARRAY_SIZE_TINY
            self.bit_array_size_bits = BIT64_ARRAY_SIZE_TINY * 64
        elif expected_records <= 62_000:
            self.bit64_array_size = BIT64_ARRAY_SIZE_SMALL
            self.bit_array_size_bits = BIT64_ARRAY_SIZE_SMALL * 64
        elif expected_records <= 1_000_000:
            self.bit64_array_size = BIT64_ARRAY_SIZE_LARGE
            self.bit_array_size_bits = BIT64_ARRAY_SIZE_LARGE * 64
        elif expected_records <= 16_000_000:
            self.bit64_array_size = BIT64_ARRAY_SIZE_HUGE
            self.bit_array_size_bits = BIT64_ARRAY_SIZE_HUGE * 64
        else:
            raise ValueError("Too many records for this Bloom filter implementation")

        # Precompute mask for faster modulo operations
        self.bit_mask = self.bit_array_size_bits - 1

        # Allocate 64-bit aligned memory
        self.bit_array = <uint64_t*>calloc(self.bit64_array_size, sizeof(uint64_t))
        if not self.bit_array:
            raise MemoryError("Failed to allocate memory for the Bloom filter.")

    def __dealloc__(self):
        if self.bit_array:
            free(self.bit_array)

    cdef inline void _add(self, const uint64_t item):
        cdef uint64_t h1, h2

        # Use bit mask for fast modulo (works because sizes are powers of 2)
        h1 = item & self.bit_mask
        # Better hash mixing for second position
        h2 = (item * GOLDEN_RATIO) & self.bit_mask

        # Set bits using 64-bit operations
        self.bit_array[h1 >> 6] |= (<uint64_t>1) << (h1 & 0x3F)
        self.bit_array[h2 >> 6] |= (<uint64_t>1) << (h2 & 0x3F)

    cpdef void add(self, const uint64_t item):
        self._add(item)

    cdef inline bint _possibly_contains(self, const uint64_t item):
        cdef uint64_t h1, h2, mask1, mask2

        h1 = item & self.bit_mask
        h2 = (item * GOLDEN_RATIO) & self.bit_mask

        # Check both bits with single 64-bit load each
        mask1 = (<uint64_t>1) << (h1 & 0x3F)
        mask2 = (<uint64_t>1) << (h2 & 0x3F)

        return (self.bit_array[h1 >> 6] & mask1) != 0 and \
               (self.bit_array[h2 >> 6] & mask2) != 0

    cpdef bint possibly_contains(self, const uint64_t item):
        return self._possibly_contains(item)

    cpdef numpy.ndarray[numpy.npy_bool, ndim=1] possibly_contains_many(self, object relation, list columns):
        """
        Optimized batch checking with better memory access patterns.
        """
        cdef Py_ssize_t num_rows = relation.num_rows
        cdef numpy.ndarray[numpy.npy_bool, ndim=1] result = numpy.zeros(num_rows, dtype=numpy.bool_)
        cdef uint8_t[::1] result_view = result
        cdef int64_t[::1] valid_row_ids = non_null_row_indices(relation, columns)
        cdef Py_ssize_t num_valid_rows = valid_row_ids.shape[0]
        cdef numpy.ndarray[numpy.uint64_t, ndim=1] row_hashes_np = numpy.zeros(num_rows, dtype=numpy.uint64)
        cdef uint64_t[::1] row_hashes = row_hashes_np
        cdef Py_ssize_t i
        cdef int64_t row_id
        cdef uint64_t hash_val, h1, h2, mask1, mask2

        if num_valid_rows == 0:
            return result

        # Compute hashes only for non-null rows
        compute_row_hashes(relation, columns, row_hashes)

        # Precompute constants
        cdef uint64_t bit_mask = self.bit_mask
        cdef uint64_t golden_ratio = GOLDEN_RATIO
        cdef uint64_t* bit_array = self.bit_array

        for i in range(num_valid_rows):
            row_id = valid_row_ids[i]
            hash_val = row_hashes[row_id]

            h1 = hash_val & bit_mask
            h2 = (hash_val * golden_ratio) & bit_mask

            mask1 = (<uint64_t>1) << (h1 & 0x3F)
            mask2 = (<uint64_t>1) << (h2 & 0x3F)

            result_view[row_id] = (bit_array[h1 >> 6] & mask1) != 0 and \
                (bit_array[h2 >> 6] & mask2) != 0

        return result

cpdef BloomFilter create_bloom_filter(object relation, list columns):
    """
    Optimized Bloom filter creation with better cache behavior.
    """
    cdef:
        Py_ssize_t num_rows = relation.num_rows
        int64_t[::1] valid_row_ids = non_null_row_indices(relation, columns)
        Py_ssize_t num_valid_rows = valid_row_ids.shape[0]
        numpy.ndarray[numpy.uint64_t, ndim=1] row_hashes_np = numpy.empty(num_rows, dtype=numpy.uint64)
        uint64_t[::1] row_hashes = row_hashes_np
        Py_ssize_t i
        int64_t row_id
        BloomFilter bf = BloomFilter(num_valid_rows)
        uint64_t hash_val, h1, h2

    if num_valid_rows == 0:
        return bf

    # Populate row hashes using the selected columns
    compute_row_hashes(relation, columns, row_hashes)

    # Precompute constants for faster access
    cdef uint64_t bit_mask = bf.bit_mask
    cdef uint64_t golden_ratio = GOLDEN_RATIO
    cdef uint64_t* bit_array = bf.bit_array

    # Add to bloom filter
    for i in range(num_valid_rows):
        row_id = valid_row_ids[i]
        hash_val = row_hashes[row_id]

        h1 = hash_val & bit_mask
        h2 = (hash_val * golden_ratio) & bit_mask

        bit_array[h1 >> 6] |= (<uint64_t>1) << (h1 & 0x3F)
        bit_array[h2 >> 6] |= (<uint64_t>1) << (h2 & 0x3F)

    return bf



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

"""
TimestampVector: Cython implementation of a fixed-width timestamp column vector for Draken.

This module provides:
- The TimestampVector class for efficient timestamp column storage (microseconds since Unix epoch)
- Integration with DrakenFixedBuffer and related C helpers for memory management
- Arrow interoperability for zero-copy conversion
- Fast comparison and null handling for timestamp columns

Used for high-performance temporal analytics and columnar data processing in Draken.
"""

from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AS_STRING
from cpython.mem cimport PyMem_Malloc
from libc.stdint cimport int32_t
from libc.stdint cimport int64_t
from libc.stdint cimport int8_t
from libc.stdint cimport intptr_t
from libc.stdint cimport uint64_t
from libc.stdint cimport uint8_t
from libc.stdlib cimport malloc
from libc.string cimport memset

from opteryx.draken.core.buffers cimport DrakenFixedBuffer
from opteryx.draken.core.buffers cimport DRAKEN_TIMESTAMP64
from opteryx.draken.core.fixed_vector cimport alloc_fixed_buffer
from opteryx.draken.core.fixed_vector cimport buf_dtype
from opteryx.draken.core.fixed_vector cimport buf_itemsize
from opteryx.draken.core.fixed_vector cimport buf_length
from opteryx.draken.core.fixed_vector cimport free_fixed_buffer
from opteryx.draken.vectors.vector cimport MIX_HASH_CONSTANT, Vector, NULL_HASH, mix_hash, simd_mix_hash


cdef inline bint _bitmap_is_valid(uint8_t* bitmap, Py_ssize_t idx, Py_ssize_t bit_offset):
    cdef Py_ssize_t bit_index = idx + bit_offset
    cdef uint8_t byte = bitmap[bit_index >> 3]
    return (byte >> (bit_index & 7)) & 1

cdef class TimestampVector(Vector):

    def __cinit__(self, size_t length=0, bint wrap=False):
        """
        length>0, wrap=False  -> allocate new owned buffer
        wrap=True             -> do not allocate; caller will set ptr & metadata
        """
        self.null_bit_offset = 0
        self._arrow_null_buf = None
        self._arrow_data_buf = None

        if wrap:
            self.ptr = NULL
            self.owns_data = False
        else:
            self.ptr = alloc_fixed_buffer(DRAKEN_TIMESTAMP64, length, 8)
            self.owns_data = True

    def __dealloc__(self):
        # Only free if we own the data and the pointer is not NULL
        if self.owns_data and self.ptr is not NULL:
            free_fixed_buffer(self.ptr, True)
            self.ptr = NULL

    # Python-friendly properties (backed by C getters for kernels)
    @property
    def length(self):
        return buf_length(self.ptr)

    @property
    def itemsize(self):
        return buf_itemsize(self.ptr)

    @property
    def dtype(self):
        return buf_dtype(self.ptr)

    def __getitem__(self, Py_ssize_t i):
        """Return the value at index i, or None if null."""
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        if i < 0 or i >= ptr.length:
            raise IndexError("Index out of bounds")
        if ptr.null_bitmap != NULL:
            if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):
                return None
        return data[i]

    # -------- Interop (owned -> Arrow) --------
    def to_arrow(self):
        """Convert to a PyArrow array."""
        import pyarrow as pa
        
        cdef size_t nbytes = buf_length(self.ptr) * buf_itemsize(self.ptr)
        addr = <intptr_t> self.ptr.data
        data_buf = pa.foreign_buffer(addr, nbytes, base=self)

        buffers = []
        cdef Py_ssize_t null_bytes
        if self.ptr.null_bitmap != NULL:
            null_bytes = (self.ptr.length + self.null_bit_offset + 7) // 8
            buffers.append(pa.foreign_buffer(<intptr_t> self.ptr.null_bitmap, null_bytes, base=self))
        else:
            buffers.append(None)

        buffers.append(data_buf)

        # Default to microsecond precision
        return pa.Array.from_buffers(pa.timestamp('us'), buf_length(self.ptr), buffers)

    # -------- Example op --------
    cpdef TimestampVector take(self, int32_t[::1] indices):
        cdef Py_ssize_t i, n = indices.shape[0]
        cdef TimestampVector out = TimestampVector(<size_t>n)
        cdef int64_t* src = <int64_t*> self.ptr.data
        cdef int64_t* dst = <int64_t*> out.ptr.data
        for i in range(n):
            dst[i] = src[indices[i]]
        return out

    cpdef int8_t[::1] equals(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] == value else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] not_equals(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] != value else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] greater_than(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] > value else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] greater_than_or_equals(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] >= value else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] less_than(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] < value else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] less_than_or_equals(self, int64_t value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        for i in range(n):
            buf[i] = 1 if data[i] <= value else 0
        return <int8_t[:n]> buf

    cpdef int64_t min(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        if n == 0:
            raise ValueError("Cannot compute min of empty column")

        cdef int64_t m
        cdef bint found = False
        cdef uint8_t byte, bit

        # Find first non-null value
        for i in range(n):
            if ptr.null_bitmap != NULL:
                if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):  # null
                    continue
            m = data[i]
            found = True
            break

        if not found:
            raise ValueError("Cannot compute min of all-null column")

        # Find minimum among remaining values
        for i in range(i + 1, n):
            if ptr.null_bitmap != NULL:
                if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):  # null
                    continue
            if data[i] < m:
                m = data[i]
        return m

    cpdef int64_t max(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        if n == 0:
            raise ValueError("Cannot compute max of empty column")

        cdef int64_t m
        cdef bint found = False
        cdef uint8_t byte, bit

        # Find first non-null value
        for i in range(n):
            if ptr.null_bitmap != NULL:
                if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):  # null
                    continue
            m = data[i]
            found = True
            break

        if not found:
            raise ValueError("Cannot compute max of all-null column")

        # Find maximum among remaining values
        for i in range(i + 1, n):
            if ptr.null_bitmap != NULL:
                if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):  # null
                    continue
            if data[i] > m:
                m = data[i]
        return m

    cpdef int8_t[::1] is_null(self):
        """
        Return a memoryview of int8_t, where each element is 1 if the value is null, 0 otherwise.
        """
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        cdef uint8_t byte, bit

        if buf == NULL:
            raise MemoryError()

        if ptr.null_bitmap == NULL:
            # No nulls — fill with 0
            for i in range(n):
                buf[i] = 0
        else:
            # Extract null bits — 1 means valid, so invert for null
            for i in range(n):
                buf[i] = 0 if _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset) else 1

        return <int8_t[:n]> buf

    @property
    def null_count(self):
        """Return the number of nulls in the vector."""
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef Py_ssize_t count = 0
        cdef uint8_t byte, bit
        if ptr.null_bitmap == NULL:
            return 0
        for i in range(n):
            if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):
                count += 1
        return count

    cpdef list to_pylist(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t i, n = ptr.length
        cdef list out = []
        cdef uint8_t byte, bit

        if ptr.null_bitmap == NULL:
            for i in range(n):
                out.append(data[i])
        else:
            for i in range(n):
                if _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):
                    out.append(data[i])
                else:
                    out.append(None)

        return out


    cdef void hash_into(
        self,
        uint64_t[::1] out_buf,
        Py_ssize_t offset=0,
    ) except *:
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef int64_t* data = <int64_t*> ptr.data
        cdef Py_ssize_t n = ptr.length

        if n == 0:
            return

        if offset < 0 or offset + n > out_buf.shape[0]:
            raise ValueError("TimestampVector.hash_into: output buffer too small")

        cdef Py_ssize_t i
        cdef uint64_t value
        cdef uint64_t* dst = &out_buf[offset]
        cdef bint has_nulls = ptr.null_bitmap != NULL
        cdef uint64_t* as_uint64 = <uint64_t*> data

        # Use shared MIX_HASH_CONSTANT directly; no need to pass it in.
        if not has_nulls:
            simd_mix_hash(dst, as_uint64, <size_t> n)
            return

        for i in range(n):
            if not _bitmap_is_valid(ptr.null_bitmap, i, self.null_bit_offset):
                value = NULL_HASH
            else:
                value = <uint64_t> data[i]

            dst[i] = mix_hash(dst[i], value)

    def __str__(self):
        cdef list vals = []
        cdef Py_ssize_t i, k = min(<Py_ssize_t>buf_length(self.ptr), 10)
        cdef int64_t* data = <int64_t*> self.ptr.data
        for i in range(k):
            vals.append(data[i])
        return f"<TimestampVector len={buf_length(self.ptr)} values={vals}>"


cdef TimestampVector from_arrow(object array):
    cdef TimestampVector vec = TimestampVector(0, True)   # wrap=True: no alloc
    vec.ptr = <DrakenFixedBuffer*> malloc(sizeof(DrakenFixedBuffer))
    if vec.ptr == NULL:
        raise MemoryError()
    vec.owns_data = False

    cdef object bufs = array.buffers()
    vec._arrow_null_buf = bufs[0]
    vec._arrow_data_buf = bufs[1]

    cdef intptr_t base_ptr = <intptr_t> bufs[1].address
    cdef size_t itemsize = 8
    cdef Py_ssize_t offset = array.offset
    cdef intptr_t nb_addr
    cdef Py_ssize_t byte_offset

    vec.ptr.type = DRAKEN_TIMESTAMP64
    vec.ptr.itemsize = itemsize
    vec.ptr.length = <size_t> len(array)

    cdef intptr_t addr = base_ptr + offset * itemsize
    vec.ptr.data = <void*> addr

    # Variables for null bitmap handling
    cdef Py_ssize_t n_bytes
    cdef bytes new_bitmap
    cdef uint8_t* dst_bitmap
    cdef uint8_t* src_bitmap
    cdef int bit_offset
    cdef int shift_down
    cdef int shift_up
    cdef uint8_t val
    cdef Py_ssize_t i

    if bufs[0] is not None:
        nb_addr = bufs[0].address
        if offset % 8 == 0:
            vec.ptr.null_bitmap = <uint8_t*> (nb_addr + (offset >> 3))
            vec.null_bit_offset = 0
        else:
            # Unaligned offset: copy and shift
            n_bytes = (vec.ptr.length + 7) // 8
            new_bitmap = PyBytes_FromStringAndSize(NULL, n_bytes)
            dst_bitmap = <uint8_t*> PyBytes_AS_STRING(new_bitmap)
            
            byte_offset = offset >> 3
            bit_offset = offset & 7
            src_bitmap = <uint8_t*> nb_addr + byte_offset
            
            shift_down = bit_offset
            shift_up = 8 - bit_offset
            
            for i in range(n_bytes):
                val = src_bitmap[i] >> shift_down
                val |= (src_bitmap[i+1] << shift_up)
                dst_bitmap[i] = val
                
            vec.ptr.null_bitmap = dst_bitmap
            vec._arrow_null_buf = new_bitmap # Keep alive
            vec.null_bit_offset = 0
    else:
        vec.ptr.null_bitmap = NULL
        vec.null_bit_offset = 0

    return vec



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Data Source Connectors with Lazy Loading

This module provides connectors to various data sources, enabling Opteryx to query
data from files, databases, cloud storage, and other systems. Connectors are lazily
loaded to improve startup performance and reduce memory footprint.

Architecture:
Connectors abstract different data sources behind a common interface (BaseConnector),
allowing the query engine to work with any data source transparently. Each connector
is responsible for:
- Reading data and converting it to PyArrow format
- Providing schema information
- Supporting predicate pushdown when possible
- Handling authentication and connection management

Connector Types:

File System Connectors:
- DiskConnector: Local file system access
- AwsS3Connector: Amazon S3 storage
- GcpCloudStorageConnector: Google Cloud Storage

Database Connectors:
- SqlConnector: SQL databases (PostgreSQL, MySQL, SQLite, etc.)
- MongoDbConnector: MongoDB document database
- CqlConnector: Cassandra/ScyllaDB

Format Connectors:
- ArrowConnector: Apache Arrow tables and datasets
- FileConnector: Generic file format handler
- IcebergConnector: Apache Iceberg tables

Special Connectors:
- VirtualDataConnector: In-memory datasets and computed tables
- InformationSchemaConnector: System metadata tables

Lazy Loading:
Connectors are only imported when actually needed, which significantly improves
module import time. The lazy loading is transparent to users - all import patterns
work normally, but the actual connector classes are loaded on first access.

Usage Patterns:

1. Direct Import:
   from opteryx.connectors import ArrowConnector

2. Registration:
   opteryx.register_store("my_prefix", my_connector_instance)

3. Query Usage:
   opteryx.query("SELECT * FROM s3://bucket/file.parquet")

Connector Development:
1. Inherit from BaseConnector
2. Implement required methods (read_dataset, get_dataset_schema)
3. Add optional optimizations (predicate pushdown, column pruning)
4. Register with appropriate prefixes
5. Add comprehensive tests

Example Custom Connector:
    class MyConnector(BaseConnector):
        def read_dataset(self, dataset, **kwargs):
            # Read data and return PyArrow table
            return pa.table(data)

        def get_dataset_schema(self, dataset):
            # Return schema information
            return pa.schema([...])

Performance Considerations:
- Implement predicate pushdown to reduce data transfer
- Support column pruning for wide tables
- Use async operations for I/O bound connectors
- Cache schema information when appropriate
- Consider connection pooling for database connectors

The lazy loading system maps prefixes to connector classes and loads them
on demand, significantly reducing initial import time while maintaining
full functionality.
"""

import os

# Lazy imports - connectors are only loaded when actually needed
# This significantly improves module import time from ~500ms to ~130ms

# load the base set of prefixes
# fmt:off
_storage_prefixes = {
    "information_schema": "InformationSchema",
    "gs:": {"connector": "GcpCloudStorageConnector", "prefix": "gs://", "remove_prefix": True},
    "s3:": {"connector": "AwsS3Connector", "prefix": "s3://", "remove_prefix": True},
    "minio:": {"connector": "AwsS3Connector", "prefix": "minio://", "remove_prefix": True},
    "file:": {"connector": "DiskConnector", "prefix": "file://", "remove_prefix": True},
}
# fmt:on

__all__ = (
    "ArrowConnector",
    "AwsS3Connector",
    "CqlConnector",
    "DiskConnector",
    "FileConnector",
    "GcpCloudStorageConnector",
    "GcpFireStoreConnector",
    "IcebergConnector",
    "MongoDbConnector",
    "SqlConnector",
)


def _lazy_import_connector(connector_name: str):
    """
    Lazy import a connector class by name.

    This function is called by __getattr__ when a connector is accessed,
    or by connector_factory when a connector needs to be instantiated.

    Args:
        connector_name: The name of the connector class to import

    Returns:
        The connector class

    Raises:
        ValueError: If the connector name is unknown
    """
    if connector_name == "ArrowConnector":
        from opteryx.connectors.arrow_connector import ArrowConnector

        return ArrowConnector
    elif connector_name == "AwsS3Connector":
        from opteryx.connectors.aws_s3_connector import AwsS3Connector

        return AwsS3Connector
    elif connector_name == "CqlConnector":
        from opteryx.connectors.cql_connector import CqlConnector

        return CqlConnector
    elif connector_name == "DiskConnector":
        from opteryx.connectors.disk_connector import DiskConnector

        return DiskConnector
    elif connector_name == "FileConnector":
        from opteryx.connectors.file_connector import FileConnector

        return FileConnector
    elif connector_name == "GcpCloudStorageConnector":
        from opteryx.connectors.gcp_cloudstorage_connector import GcpCloudStorageConnector

        return GcpCloudStorageConnector
    elif connector_name == "GcpFireStoreConnector":
        from opteryx.connectors.gcp_firestore_connector import GcpFireStoreConnector

        return GcpFireStoreConnector
    elif connector_name == "IcebergConnector":
        from opteryx.connectors.iceberg_connector import IcebergConnector

        return IcebergConnector
    elif connector_name == "MongoDbConnector":
        from opteryx.connectors.mongodb_connector import MongoDbConnector

        return MongoDbConnector
    elif connector_name == "SqlConnector":
        from opteryx.connectors.sql_connector import SqlConnector

        return SqlConnector
    else:
        raise ValueError(f"Unknown connector: {connector_name}")


def register_store(prefix, connector, *, remove_prefix: bool = False, **kwargs):
    """add a prefix"""
    if not isinstance(connector, type):  # type: ignore
        # uninstantiated classes aren't a type
        raise ValueError("connectors registered with `register_store` must be uninstantiated.")

    if connector.__name__ == "IcebergConnector" and not remove_prefix:
        raise ValueError(
            "IcebergConnector requires remove_prefix=True so catalog prefixes don't leak into table names."
        )

    # Store connector class directly (not as a string)
    _storage_prefixes[prefix] = {
        "connector": connector,  # type: ignore
        "prefix": prefix,
        "remove_prefix": remove_prefix,
        **kwargs,
    }


def register_df(name, frame):
    """register a orso, pandas or Polars dataframe"""
    # Lazy import ArrowConnector
    import pyarrow

    from opteryx.connectors.arrow_connector import ArrowConnector

    # polars (maybe others) - the polars to arrow API is a mess
    if hasattr(frame, "_df"):  # pragma: no cover
        frame = frame._df
    if "PyDataFrame" in str(type(frame)):  # pragma: no cover
        arrow = frame.to_arrow(compat_level=1)
        if not isinstance(arrow, pyarrow.Table):
            from opteryx.exceptions import NotSupportedError

            raise NotSupportedError(
                "Polars not supported whilst changes are being made to the Polars to Arrow APIs"
            )
        register_arrow(name, arrow)
        return
    if hasattr(frame, "to_arrow"):  # pragma: no cover
        arrow = frame.to_arrow()
        if not isinstance(arrow, pyarrow.Table):
            arrow = pyarrow.Table.from_batches(arrow)
        register_arrow(name, arrow)
        return
    # orso
    if hasattr(frame, "arrow"):  # pragma: no cover
        arrow = frame.arrow()
        register_arrow(name, arrow)
        return
    # pandas
    frame_type = str(type(frame))
    if "pandas" in frame_type:  # pragma: no cover
        register_arrow(name, pyarrow.Table.from_pandas(frame))
        return
    raise ValueError("Unable to register unknown frame type.")


def register_arrow(name, table):
    """register an arrow table"""
    from opteryx.connectors.arrow_connector import ArrowConnector
    from opteryx.shared import MaterializedDatasets

    materialized_datasets = MaterializedDatasets()
    materialized_datasets[name] = table
    register_store(name, ArrowConnector)


def known_prefix(prefix) -> bool:
    return prefix in _storage_prefixes


def connector_factory(dataset, statistics, **config):
    """
    Work out which connector will service the access to this dataset.
    """

    # if it starts with a $, it's a special internal dataset
    if dataset[0] == "$":
        from opteryx.connectors import virtual_data

        return virtual_data.SampleDataConnector(dataset=dataset, statistics=statistics)

    # Look up the prefix from the registered prefixes
    connector_entry: dict = config
    for prefix, storage_details in _storage_prefixes.items():
        if (
            dataset == prefix
            or dataset.startswith(prefix + ".")
            or dataset.startswith(prefix + "//")
        ):
            connector_entry.update(storage_details.copy())  # type: ignore
            connector = connector_entry.pop("connector")
            # If connector is a string, lazy load it
            if isinstance(connector, str):
                connector = _lazy_import_connector(connector)
            break
    else:
        # Check if dataset is a file or contains wildcards
        has_wildcards = any(char in dataset for char in ["*", "?", "["])
        if os.path.isfile(dataset) or has_wildcards:
            from opteryx.connectors import file_connector

            return file_connector.FileConnector(dataset=dataset, statistics=statistics)
        # fall back to the default connector (local disk if not set)
        connector_entry = _storage_prefixes.get("_default", {})
        connector = connector_entry.pop("connector", "DiskConnector")
        # If connector is a string, lazy load it
        if isinstance(connector, str):
            connector = _lazy_import_connector(connector)

    prefix = connector_entry.pop("prefix", "")
    remove_prefix = connector_entry.pop("remove_prefix", False)
    if prefix and remove_prefix and dataset.startswith(prefix):
        # Remove the prefix. If there's a separator (. or //) after the prefix, skip it too
        dataset = dataset[len(prefix) :]
        if dataset.startswith(".") or dataset.startswith("//"):
            dataset = dataset[1:] if dataset.startswith(".") else dataset[2:]
        if connector.__name__ == "IcebergConnector" and dataset and "." not in dataset:
            # Default to a namespace matching the prefix when one isn't provided
            dataset = f"{prefix}.{dataset}"

    return connector(dataset=dataset, statistics=statistics, **connector_entry)


def __getattr__(name):
    """
    Lazy load connector classes when accessed as module attributes.

    This allows the standard import pattern to work:
        from opteryx.connectors import ArrowConnector

    But defers the actual import until the connector is accessed,
    significantly improving initial module load time.

    Args:
        name: The attribute name being accessed

    Returns:
        The connector class if it exists in __all__

    Raises:
        AttributeError: If the attribute doesn't exist
    """
    if name in __all__:
        return _lazy_import_connector(name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
This is a virtual dataset which is calculated at access time.

It is the system variables collection.
"""

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx.models import RelationStatistics

__all__ = ("read", "schema")


def read(end_date=None, variables={}):
    import pyarrow

    buffer = []
    for variable in variables:
        variable_type, variable_value, variable_owner, variable_visibility = variables.details(
            variable
        )
        buffer.append(
            {
                "name": variable,
                "value": str(variable_value),
                "type": variable_type,
                "owner": variable_owner.name,
                "visibility": variable_visibility.name,
            }
        )

    return pyarrow.Table.from_pylist(buffer)


def schema():
    # fmt:off
    return  RelationSchema(
        name="$variables",
        columns=[
            FlatColumn(name="name", type=OrsoTypes.VARCHAR),
            FlatColumn(name="value", type=OrsoTypes.VARCHAR),
            FlatColumn(name="type", type=OrsoTypes.VARCHAR),
            FlatColumn(name="owner", type=OrsoTypes.VARCHAR),
            FlatColumn(name="visibility", type=OrsoTypes.VARCHAR),
        ],
    )
    # fmt:on


def statistics() -> RelationStatistics:
    return RelationStatistics()



"""
GCS-backed Key-Value Store

Expects a location like: gs://bucket/[optional-prefix]
The key provided will be the filename portion of the object key.
"""

from __future__ import annotations

import importlib
from typing import Iterable
from typing import Union
from urllib.parse import urlparse

from orso.tools import single_item_cache

from opteryx.exceptions import MissingDependencyError
from opteryx.managers.kvstores import BaseKeyValueStore

GoogleAPIError = Exception


@single_item_cache
def _gcs_client(**_kwargs):
    try:
        storage = importlib.import_module("google.cloud.storage")
    except ImportError as err:  # pragma: no cover - optional dependency
        raise MissingDependencyError("google-cloud-storage") from err
    try:
        global GoogleAPIError
        GoogleAPIError = importlib.import_module("google.api_core.exceptions").GoogleAPIError
    except Exception:
        GoogleAPIError = Exception
    return storage.Client()


class GCSKeyValueStore(BaseKeyValueStore):
    def __init__(self, location: str, **_kwargs):
        parsed = urlparse(location)
        if parsed.scheme != "gs":
            raise ValueError("location must be a gs:// URI")

        self._bucket_name = parsed.netloc
        self._prefix = parsed.path.lstrip("/")
        self._client = _gcs_client(**_kwargs)
        self._bucket = self._client.bucket(self._bucket_name)
        super().__init__(location)

    def _object_name(self, key: bytes) -> str:
        try:
            key_str = key.decode("utf-8")
        except UnicodeDecodeError:
            key_str = key.hex()
        if self._prefix:
            return f"{self._prefix}/{key_str}"
        return key_str

    def get(self, key: bytes) -> Union[bytes, None]:
        name = self._object_name(key)
        blob = self._bucket.blob(name)
        try:
            return blob.download_as_bytes()
        except (GoogleAPIError, KeyError):
            return None

    def set(self, key: bytes, value: bytes) -> None:
        name = self._object_name(key)
        blob = self._bucket.blob(name)
        blob.upload_from_string(value)

    def contains(self, keys: Iterable) -> Iterable:
        result = []
        for k in keys:
            blob = self._bucket.blob(self._object_name(k))
            try:
                if blob.exists():
                    result.append(k)
            except (GoogleAPIError, KeyError):
                continue
        return result

    def delete(self, key: bytes) -> None:
        blob = self._bucket.blob(self._object_name(key))
        try:
            blob.delete()
        except (GoogleAPIError, KeyError):
            pass



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from libc.stdint cimport uint64_t

from opteryx.draken.vectors.vector cimport Vector


cpdef void hash_into(
    Vector vector,
    uint64_t[::1] out_buf,
    Py_ssize_t offset=0
):
    """
    Python-visible shim for invoking Vector.hash_into.

    This exists solely for test helpers; production code should call the cdef
    method directly from Cython.
    """
    vector.hash_into(out_buf, offset)



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from libc.stddef cimport size_t
from libc.stdint cimport uint64_t

from opteryx.draken.interop.arrow import vector_from_arrow

cdef const uint64_t MIX_HASH_CONSTANT
cdef const uint64_t NULL_HASH

cdef extern from "simd_hash.h":
    void simd_mix_hash(uint64_t* dest, const uint64_t* values, size_t count) nogil

cdef inline uint64_t mix_hash(uint64_t current, uint64_t value) nogil:
    cdef uint64_t mixed = current ^ value
    mixed = mixed * MIX_HASH_CONSTANT + 1
    return mixed ^ (mixed >> 32)

cdef class Vector:
    cdef bint here
    cpdef object null_bitmap(self)
    cdef void hash_into(self, uint64_t[::1] out_buf, Py_ssize_t offset=*) except *



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from cython import Py_ssize_t
from libc.stdint cimport int64_t, uint8_t, int32_t

import numpy
cimport numpy
numpy.import_array()

from opteryx.draken.vectors.vector cimport Vector
from opteryx.draken.vectors.int64_vector cimport Int64Vector
from opteryx.draken.vectors.string_vector cimport StringVector
from opteryx.draken.core.buffers cimport DrakenVarBuffer

cdef numpy.ndarray[numpy.uint8_t, ndim=1] list_in_list_int64_vector(Int64Vector vec, set values):
    cdef Py_ssize_t i, size = vec.ptr.length
    cdef numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.empty(size, dtype=numpy.uint8)
    cdef uint8_t[::1] result_view = result
    cdef int64_t* data = <int64_t*>vec.ptr.data
    cdef uint8_t* nulls = vec.ptr.null_bitmap
    cdef bint null_in_values = None in values
    cdef bint is_valid

    if nulls == NULL:
        for i in range(size):
            result_view[i] = data[i] in values
    else:
        for i in range(size):
            is_valid = (nulls[i >> 3] >> (i & 7)) & 1
            if is_valid:
                result_view[i] = data[i] in values
            else:
                result_view[i] = null_in_values

    return result

cdef numpy.ndarray[numpy.uint8_t, ndim=1] list_in_list_string_vector(StringVector vec, set values):
    cdef Py_ssize_t i, size = vec.ptr.length
    cdef numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.empty(size, dtype=numpy.uint8)
    cdef uint8_t[::1] result_view = result

    # Convert string values to bytes for comparison
    cdef set bytes_values = set()
    for val in values:
        if val is None:
            bytes_values.add(None)
        elif isinstance(val, bytes):
            bytes_values.add(val)
        elif isinstance(val, str):
            bytes_values.add(val.encode('utf-8'))
        else:
            bytes_values.add(val)

    cdef DrakenVarBuffer* ptr = vec.ptr
    cdef int32_t start, end
    cdef char* data = <char*>ptr.data
    cdef uint8_t* nulls = ptr.null_bitmap
    cdef bint null_in_values = None in bytes_values
    cdef bint is_valid
    cdef bytes s

    if nulls == NULL:
        for i in range(size):
            start = ptr.offsets[i]
            end = ptr.offsets[i+1]
            s = data[start:end]
            result_view[i] = s in bytes_values
    else:
        for i in range(size):
            is_valid = (nulls[i >> 3] >> (i & 7)) & 1
            if is_valid:
                start = ptr.offsets[i]
                end = ptr.offsets[i+1]
                s = data[start:end]
                result_view[i] = s in bytes_values
            else:
                result_view[i] = null_in_values

    return result

cpdef numpy.ndarray[numpy.uint8_t, ndim=1] list_in_list(object arr, set values):
    """
    Fast membership check for "InList" using Cython.

    Parameters:
        arr: NumPy array or Draken Vector.
        values: Set of valid values.

    Returns:
        NumPy boolean array indicating membership.
    """
    # Handle Draken vectors
    if isinstance(arr, Int64Vector):
        return list_in_list_int64_vector(arr, values)
    elif isinstance(arr, StringVector):
        return list_in_list_string_vector(arr, values)
    elif isinstance(arr, Vector):
        # Fallback for other vector types - convert to pylist and use numpy path
        arr = numpy.array(arr.to_pylist(), dtype=object)

    # Handle NumPy arrays
    cdef Py_ssize_t i, size = arr.shape[0]
    cdef numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.empty(size, dtype=numpy.uint8)
    cdef uint8_t[::1] result_view = result
    cdef object[::1] arr_view = arr

    for i in range(size):
        result_view[i] = arr_view[i] in values

    return result

cpdef numpy.ndarray[numpy.uint8_t, ndim=1] list_in_list_int64(const int64_t[::1] arr, set values, Py_ssize_t size):
    """
    Fast membership check for "InList" using Cython.

    Parameters:
        arr: NumPy array of arbitrary type (should be homogeneous).
        values: List of valid values (converted to a Cython set).

    Returns:
        NumPy boolean array indicating membership.
    """
    cdef Py_ssize_t i
    cdef numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.empty(size, dtype=numpy.uint8)
    cdef uint8_t[::1] result_view = result
    cdef int64_t value

    for i in range(size):
        value = arr[i]
        result_view[i] = value in values

    return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

import datetime

import numpy
import pyarrow
from pyarrow import compute

from opteryx.exceptions import InvalidFunctionParameterError
from opteryx.exceptions import InvalidInternalStateError
from opteryx.exceptions import SqlError
from opteryx.utils.dates import parse_iso


def convert_int64_array_to_pyarrow_datetime(values: numpy.ndarray) -> pyarrow.Array:
    """
    Convert a NumPy int64 array to PyArrow TimestampArray, inferring time unit.
    """
    if isinstance(values, pyarrow.ChunkedArray):
        values = values.to_numpy(zero_copy_only=False)

    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(zero_copy_only=False)

    if not isinstance(values, numpy.ndarray):
        raise InvalidInternalStateError("Expected a NumPy int64 array.")

    if not numpy.issubdtype(values.dtype, numpy.integer):
        raise ValueError("Cannot convert non-integer array to a timestamp.")

    min_value = values.min()
    max_value = values.max()

    # Range dispatch table: (low, high, datetime unit)
    RANGES = [
        (1e0, 1e6, "D"),
        (1e9, 1e10, "s"),
        (1e12, 1e13, "ms"),
        (1e15, 1e16, "us"),
        (1e18, 1e19, "ns"),
    ]

    for low, high, unit in RANGES:
        if low <= min_value < high and low <= max_value < high:
            try:
                return pyarrow.array(values.astype(f"datetime64[{unit}]"))
            except Exception as e:
                raise ValueError(f"Failed to cast to datetime64[{unit}]: {e}")

    raise ValueError(
        f"Unable to determine timestamp precision for values in range [{min_value}, {max_value}]"
    )


def date_part(part, arr):
    """
    Also the EXTRACT function - we extract a given part from an array of dates
    """
    j2000_scalar = numpy.array([numpy.datetime64("2000-01-01T12:00:00", "us")])
    extractors = {
        "nanosecond": compute.nanosecond,
        "nanoseconds": compute.nanosecond,
        "microsecond": compute.microsecond,
        "microseconds": compute.microsecond,
        "millisecond": compute.millisecond,
        "milliseconds": compute.millisecond,
        "second": compute.second,
        "minute": compute.minute,
        "hour": compute.hour,
        "day": compute.day,
        "dayofweek": compute.day_of_week,
        "dow": compute.day_of_week,
        "date": lambda x: compute.cast(x, "date32"),
        "week": compute.week,
        "isoweek": compute.iso_week,
        "month": compute.month,
        "quarter": compute.quarter,
        "dayofyear": compute.day_of_year,
        "doy": compute.day_of_year,
        "year": compute.year,
        "isoyear": compute.iso_year,
        "decade": lambda x: compute.divide(compute.year(x), 10),
        "century": lambda x: compute.add(compute.divide(compute.year(x), 100), 1),
        "epoch": lambda x: compute.divide(compute.cast(x, "int64"), 1000000.00),
        "julian": lambda x: compute.add(
            compute.divide(compute.milliseconds_between(x, j2000_scalar), 86400000.0),
            2451545.0,
        ),
        # ** supported by parser but not by pyarrow
        # isodow
        # millenium
        # millennium
        # timezone
        # time
        # timezonehour
        # timezoneminute
    }

    # if we get a date literal
    if not hasattr(arr, "__iter__"):
        arr = numpy.array([arr])

    arr = numpy.array(arr, dtype="datetime64[us]")

    part = part[0].lower()  # [#325]
    if part in extractors:
        return extractors[part](arr)

    from opteryx.utils import suggest_alternative

    alt = suggest_alternative(part, list(extractors.keys()))
    if not alt:
        raise InvalidFunctionParameterError(
            f"Date part `{part}` unsupported for EXTRACT."
        )  # pragma: no cover
    raise InvalidFunctionParameterError(
        f"Date part `{part}` unsupported for EXTRACT. Did you mean '{alt}'?"
    )


def date_diff(part, start, end):
    """calculate the difference between timestamps"""

    from opteryx.compiled.list_ops import list_date_diff

    arrow_extractors = {
        "months": compute.month_interval_between,  # this one doesn't work
        "quarters": compute.quarters_between,
        "weeks": compute.weeks_between,
        "years": compute.years_between,
    }

    part = str(part[0]).lower()
    if not part.endswith("s"):
        part += "s"

    if part in arrow_extractors:
        diff = arrow_extractors[part](start, end)
        if not hasattr(diff, "__iter__"):
            diff = numpy.array([diff])
        return [i.as_py() for i in diff]

    # --- normalize `start` ---
    if isinstance(start, numpy.ndarray):
        if start.dtype == numpy.int64:
            pass  # already correct
        elif numpy.issubdtype(start.dtype, numpy.datetime64):
            start = start.astype("datetime64[us]").astype(numpy.int64)
        else:
            # likely dtype=object → datetimes, convert explicitly
            start = numpy.array(
                [
                    int(x.timestamp() * 1_000_000) if isinstance(x, datetime.datetime) else 0
                    for x in start
                ],
                dtype=numpy.int64,
            )

    elif isinstance(start, pyarrow.Array):
        start = start.cast("timestamp[us]").to_numpy()
    else:
        # scalar or iterable
        start = numpy.array(
            [int(x.timestamp() * 1_000_000) for x in numpy.atleast_1d(start)],
            dtype=numpy.int64,
        )

    # --- normalize `end` ---
    if isinstance(end, numpy.ndarray):
        if end.dtype == numpy.int64:
            pass
        elif numpy.issubdtype(end.dtype, numpy.datetime64):
            end = end.astype("datetime64[us]").astype(numpy.int64)
        else:
            end = numpy.array(
                [
                    int(x.timestamp() * 1_000_000) if isinstance(x, datetime.datetime) else 0
                    for x in end
                ],
                dtype=numpy.int64,
            )

    elif isinstance(end, pyarrow.Array):
        end = end.cast("timestamp[us]").to_numpy()
    else:
        end = numpy.array(
            [int(x.timestamp() * 1_000_000) for x in numpy.atleast_1d(end)],
            dtype=numpy.int64,
        )

    return list_date_diff(start, end, part)


def time_diff(time1, time2):
    return date_diff(["hours"], time1, time2)


def date_format(dates, pattern):  # [#325]
    pattern = pattern[0]
    return [None if d is None else d.strftime(pattern) for d in dates.tolist()]


def date_floor(dates, magnitude, units):  # [#325]
    return compute.floor_temporal(dates, magnitude[0], units[0])


def from_unixtimestamp(values):
    return numpy.array(
        [datetime.datetime.fromtimestamp(i, tz=datetime.timezone.utc) for i in values],
        dtype="datetime64[s]",
    )


def unixtime(array):
    """
    Convert a NumPy or Arrow array of timestamps or ISO8601 strings to Unix time (seconds since epoch).
    NaNs or nulls are converted to numpy.nan.
    """
    if isinstance(array, pyarrow.ChunkedArray):
        # Handle ChunkedArray by processing chunks individually
        if array.num_chunks == 0:
            return numpy.array([], dtype=numpy.int64)
        chunks = [unixtime(chunk) for chunk in array.chunks]
        return numpy.concatenate(chunks)

    if numpy.issubdtype(array.dtype, numpy.datetime64):
        # Convert datetime64[ns] to seconds since epoch
        return array.astype("datetime64[s]").astype(numpy.int64)

    elif array.dtype.kind in {"U", "S", "O"}:
        # Assume strings: parse to datetime and convert
        def to_epoch(s):
            if s is None or s != s:
                return numpy.datetime64("NaT")
            try:
                dt = numpy.datetime64(s, "s")
                return dt.astype(numpy.int64)
            except Exception:
                return numpy.datetime64("NaT")

        return numpy.vectorize(to_epoch)(array)

    else:
        raise TypeError(f"Unsupported array type: {array.dtype}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
This is the async wrapper around the memory pool, it is used by the
async blob readers to move data to the synchronous code.
"""

import asyncio

from opteryx.compiled.structures.memory_pool import MemoryPool


class AsyncMemoryPool:
    def __init__(self, pool: MemoryPool):
        self.pool: MemoryPool = pool
        # MemoryPool is a blocking, thread-safe object (uses an RLock). We avoid
        # serialising all async operations on a single asyncio.Lock which causes
        # the event loop to become a bottleneck when many coroutines call
        # commit/read/release concurrently. Instead we run the blocking calls in
        # the default thread pool executor. This keeps the event loop responsive
        # while still using the compiled MemoryPool which provides C-level
        # performance for the actual memory operations.
        self.lock = None

    async def commit(self, data: bytes) -> int:
        loop = asyncio.get_running_loop()
        # Offload the blocking commit to a thread to avoid blocking the event loop
        return await loop.run_in_executor(None, self.pool.commit, data)

    async def read(self, ref_id: int, zero_copy=True, latch=True) -> bytes:
        """
        In an async environment, we much more certain the bytes will be overwritten
        if we don't materialize them so we always create a copy.
        """
        loop = asyncio.get_running_loop()
        # Offload the blocking read to a thread. The compiled MemoryPool.read
        # is fast but still uses locks and may block; moving it to a thread
        # allows multiple reads/commits to progress concurrently.
        return await loop.run_in_executor(None, self.pool.read, ref_id, zero_copy, latch)

    async def release(self, ref_id: int):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.pool.release, ref_id)

    def size(self):
        return self.pool.size



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import numpy as np
cimport numpy as cnp
cnp.import_array()
from libc.stdint cimport int64_t


cdef inline int64_t min3(int64_t x, int64_t y, int64_t z) nogil:
    """Utility function to find the minimum of three integers."""
    if x <= y:
        if x <= z:
            return x
        return z
    if y <= z:
        return y
    return z


cdef inline int64_t levenshtein_single(str string1, str string2):
    """
    Calculate the Levenshtein distance between two strings.

    Parameters:
        string1 (str): The first string to compare.
        string2 (str): The second string to compare.

    Returns:
        int: The Levenshtein distance between string1 and string2.
    """
    if len(string1) < len(string2):
        string1, string2 = string2, string1

    cdef int len1 = len(string1)
    cdef int len2 = len(string2) + 1

    cdef int64_t i, j

    # Allocate a numpy array and create a memory view from it
    cdef int64_t[:] dp = np.zeros((len1 + 1) * len2, dtype=np.int64)

    for i in range(len1 + 1):
        for j in range(len2):
            if i == 0:
                dp[j] = j
            elif j == 0:
                dp[i * len2] = i
            elif string1[i - 1] == string2[j - 1]:
                dp[i * len2 + j] = dp[(i - 1) * len2 + (j - 1)]
            else:
                dp[i * len2 + j] = 1 + min3(
                    dp[(i - 1) * len2 + j],  # Remove
                    dp[i * len2 + (j - 1)],  # Insert
                    dp[(i - 1) * len2 + (j - 1)]  # Replace
                )

    return dp[len1 * len2 + (len2 - 1)]


cpdef cnp.ndarray[cnp.int64_t, ndim=1] list_levenshtein(cnp.ndarray[object, ndim=1] a, cnp.ndarray[object, ndim=1] b):
    """
    Calculate Levenshtein distance for arrays of strings.

    Parameters:
        a: Array of strings
        b: Array of strings

    Returns:
        Array of integers representing Levenshtein distances
    """
    cdef Py_ssize_t size = len(a)
    cdef cnp.ndarray[cnp.int64_t, ndim=1] result = np.zeros(size, dtype=np.int64)
    cdef Py_ssize_t i

    for i in range(size):
        if a[i] is None or b[i] is None:
            result[i] = -1  # or some other value to indicate null
        else:
            result[i] = levenshtein_single(str(a[i]), str(b[i]))

    return result



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from cpython.buffer cimport PyBUF_CONTIG_RO, PyBuffer_Release, PyObject_GetBuffer
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cpython.unicode cimport PyUnicode_FromString
from libc.stddef cimport size_t
from libc.stdint cimport uint8_t


cdef extern from "zstd.h":
    size_t ZSTD_decompressBound(const void* src, size_t srcSize)
    size_t ZSTD_decompress(void* dst, size_t dstCapacity, const void* src, size_t srcSize)
    bint ZSTD_isError(size_t code)
    const char* ZSTD_getErrorName(size_t code)
    unsigned long long ZSTD_findDecompressedSize(const void* src, size_t srcSize)
    unsigned long long ZSTD_CONTENTSIZE_ERROR
    unsigned long long ZSTD_CONTENTSIZE_UNKNOWN

cpdef bytes decompress(object data):
    """Decompress a buffer compressed with the vendored zstd sources."""
    cdef Py_buffer view
    cdef size_t dest_capacity
    cdef const uint8_t* src
    cdef size_t src_size
    cdef unsigned long long known_size
    cdef size_t result
    cdef void* dst

    if PyObject_GetBuffer(data, &view, PyBUF_CONTIG_RO) != 0:
        raise TypeError("expected a buffer-like object")

    try:
        if view.len == 0:
            return b""

        src = <const uint8_t*>view.buf
        src_size = <size_t>view.len

        known_size = ZSTD_findDecompressedSize(src, src_size)
        if known_size not in (ZSTD_CONTENTSIZE_ERROR, ZSTD_CONTENTSIZE_UNKNOWN):
            dest_capacity = <size_t>known_size
        else:
            dest_capacity = ZSTD_decompressBound(src, src_size)
            if dest_capacity == 0 or ZSTD_isError(dest_capacity):
                raise ValueError("failed to determine vendored zstd decompress bound")

        if dest_capacity == 0:
            dest_capacity = 1

        dst = PyMem_Malloc(dest_capacity)
        if not dst:
            raise MemoryError()

        result = ZSTD_decompress(dst, dest_capacity, src, src_size)
        if ZSTD_isError(result):
            err_ptr = ZSTD_getErrorName(result)
            PyMem_Free(dst)
            if err_ptr:
                raise ValueError(PyUnicode_FromString(err_ptr))
            raise ValueError("zstd decompression failed")

        py_bytes = PyBytes_FromStringAndSize(<char*>dst, result)
        PyMem_Free(dst)
        if py_bytes is None:
            raise MemoryError()
        return py_bytes
    finally:
        PyBuffer_Release(&view)



from .boolean_simplication import BooleanSimplificationStrategy
from .constant_folding import ConstantFoldingStrategy
from .correlated_filters import CorrelatedFiltersStrategy
from .distinct_pushdown import DistinctPushdownStrategy
from .join_ordering import JoinOrderingStrategy
from .join_rewriter import JoinRewriteStrategy
from .limit_pushdown import LimitPushdownStrategy
from .operator_fusion import OperatorFusionStrategy
from .predicate_compaction import PredicateCompactionStrategy
from .predicate_ordering import PredicateOrderingStrategy
from .predicate_pushdown import PredicatePushdownStrategy
from .predicate_rewriter import PredicateRewriteStrategy
from .projection_pushdown import ProjectionPushdownStrategy
from .redundant_operators import RedundantOperationsStrategy
from .split_conjunctive_predicates import SplitConjunctivePredicatesStrategy

__all__ = [
    "BooleanSimplificationStrategy",
    "ConstantFoldingStrategy",
    "CorrelatedFiltersStrategy",
    "DistinctPushdownStrategy",
    "JoinOrderingStrategy",
    "JoinRewriteStrategy",
    "LimitPushdownStrategy",
    "OperatorFusionStrategy",
    "PredicateCompactionStrategy",
    "PredicateOrderingStrategy",
    "PredicatePushdownStrategy",
    "PredicateRewriteStrategy",
    "ProjectionPushdownStrategy",
    "RedundantOperationsStrategy",
    "SplitConjunctivePredicatesStrategy",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Unnest Join Node

This is a SQL Query Execution Plan Node.

This implements a CROSS JOIN UNNEST, this isn't really a JOIN in that it doesn't join two tables
together, but it does unnest a column in a table and repeat the rows in the table for each value.
"""

from typing import Generator
from typing import Set
from typing import Tuple

import numpy
import pyarrow
from orso.schema import FlatColumn

from opteryx import EOS
from opteryx.managers.expression import NodeType
from opteryx.models import LogicalColumn
from opteryx.models import QueryProperties
from opteryx.third_party.abseil.containers import FlatHashSet

from . import BasePlanNode

INTERNAL_BATCH_SIZE: int = 10000  # config
MAX_JOIN_SIZE: int = 1000  # config
MORSEL_SIZE_BYTES: int = 16 * 1024 * 1024
CROSS_JOIN_UNNEST_BATCH_SIZE = 10000


def _cross_join_unnest_column(
    *,
    morsel: pyarrow.Table = None,
    source: LogicalColumn = None,
    target_column: FlatColumn = None,
    conditions: Set = None,
    distinct: bool = False,
    single_column: bool = False,
    hash_set=None,
) -> pyarrow.Table:
    """
    Perform a cross join on an unnested column of pyarrow tables.

    Args:
        morsels: An iterable of `pyarrow.Table` objects to be cross joined.
        source: The source node indicating the column.
        target_column: The column to be unnested.

    Returns:
        A generator that yields the resulting `pyarrow.Table` objects.
    """
    from opteryx.compiled.joins import build_filtered_rows_indices_and_column
    from opteryx.compiled.joins import build_rows_indices_and_column
    from opteryx.compiled.joins import list_distinct
    from opteryx.compiled.joins import numpy_build_filtered_rows_indices_and_column
    from opteryx.compiled.joins import numpy_build_rows_indices_and_column

    batch_size: int = INTERNAL_BATCH_SIZE
    at_least_once = False
    single_column_collector = []

    # Break the morsel into batches to avoid memory issues
    for left_block in morsel.to_batches(max_chunksize=batch_size):
        new_block = None
        # Fetch the data of the column to be unnested
        column_data = left_block[source.schema_column.identity]

        # Filter out null values
        valid_offsets = column_data.is_valid()
        if valid_offsets.true_count == 0:
            continue
        column_data = column_data.filter(valid_offsets)
        left_block = left_block.filter(valid_offsets)

        # Build indices and new column data
        if conditions is None:
            if hasattr(column_data.type, "value_type") and (
                column_data.type.value_type == pyarrow.string()
                or column_data.type.value_type == pyarrow.binary()
            ):
                # optimized version for string and binary columns
                indices, new_column_data = build_rows_indices_and_column(column_data)
            else:
                # fallback to numpy version
                indices, new_column_data = numpy_build_rows_indices_and_column(
                    column_data.to_numpy(False)
                )
        else:
            if hasattr(column_data.type, "value_type") and (
                column_data.type.value_type == pyarrow.string()
                or column_data.type.value_type == pyarrow.binary()
            ):
                indices, new_column_data = build_filtered_rows_indices_and_column(
                    column_data, conditions
                )
            else:
                indices, new_column_data = numpy_build_filtered_rows_indices_and_column(
                    column_data.to_numpy(False), conditions
                )

        if single_column and distinct and indices.size > 0:
            # if the unnest target is the only field in the SELECT and we're DISTINCTING
            indices = numpy.array(indices, dtype=numpy.int64)
            new_column_data, indices, hash_set = list_distinct(new_column_data, indices, hash_set)

        if len(indices) > 0:
            if single_column:
                single_column_collector.extend(new_column_data)
                if len(single_column_collector) > INTERNAL_BATCH_SIZE:
                    schema = pyarrow.schema(
                        [
                            pyarrow.field(
                                name=target_column.identity, type=target_column.arrow_field.type
                            )
                        ]
                    )
                    arrow_array = pyarrow.array(single_column_collector)
                    if arrow_array.type != target_column.arrow_field.type:
                        arrow_array = arrow_array.cast(target_column.arrow_field.type)
                    new_block = pyarrow.Table.from_arrays([arrow_array], schema=schema)
                    single_column_collector.clear()
                    del arrow_array
                    yield new_block
                    at_least_once = True
            else:
                # Rebuild the block with the new column data if we have any rows to build for

                total_rows = indices.size  # Both arrays have the same length
                block_size = MORSEL_SIZE_BYTES / (left_block.nbytes / left_block.num_rows)
                block_size = int(block_size / 1000) * 1000

                for start_block in range(0, total_rows, block_size):
                    # Compute the end index for the current chunk
                    end_block = min(start_block + block_size, total_rows)

                    # Slice the current chunk of indices and new_column_data
                    indices_chunk = indices[start_block:end_block]
                    new_column_data_chunk = new_column_data[start_block:end_block]

                    # Create a new block using the chunk of indices
                    indices_chunk = numpy.array(indices_chunk, dtype=numpy.int64)
                    new_block = left_block.take(indices_chunk)
                    new_block = pyarrow.Table.from_batches([new_block], schema=morsel.schema)

                    # Append the corresponding chunk of new_column_data to the block
                    new_block = new_block.append_column(
                        target_column.identity, pyarrow.array(new_column_data_chunk)
                    )

                    yield new_block
                    at_least_once = True

    if single_column_collector:
        schema = pyarrow.schema(
            [pyarrow.field(name=target_column.identity, type=target_column.arrow_field.type)]
        )
        arrow_array = pyarrow.array(single_column_collector)
        if arrow_array.type != target_column.arrow_field.type:
            arrow_array = arrow_array.cast(target_column.arrow_field.type)
        new_block = pyarrow.Table.from_arrays([arrow_array], schema=schema)
        yield new_block
        at_least_once = True

    if not at_least_once:
        # Create an empty table with the new schema
        schema = morsel.schema
        new_column = pyarrow.field(target_column.identity, pyarrow.string())
        new_schema = pyarrow.schema(list(schema) + [new_column])
        new_block = pyarrow.Table.from_batches([], schema=new_schema)
        yield new_block


def _cross_join_unnest_literal(
    morsel: pyarrow.Table, source: Tuple, target_column: FlatColumn
) -> Generator[pyarrow.Table, None, None]:
    joined_list_size = len(source)

    # Break the morsel into batches to avoid memory issues
    for left_block in morsel.to_batches(max_chunksize=INTERNAL_BATCH_SIZE):
        left_block = pyarrow.Table.from_batches([left_block], schema=morsel.schema)
        block_size = left_block.num_rows

        # Repeat each row in the table n times
        repeated_indices = numpy.repeat(numpy.arange(block_size), joined_list_size)
        appended_table = left_block.take(repeated_indices)

        # Tile the array to match the new number of rows
        tiled_array = numpy.tile(source, block_size)

        # Convert tiled_array to PyArrow array and append it to the table
        array_column = pyarrow.array(tiled_array)
        appended_table = appended_table.append_column(target_column.identity, array_column)

        yield appended_table


class UnnestJoinNode(BasePlanNode):
    """
    Implements CROSS JOIN UNNEST
    """

    is_stateless = True

    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)

        # do we have unnest details?
        self._unnest_column = parameters.get("unnest_column")
        self._unnest_target = parameters.get("unnest_target").schema_column
        self._filters = parameters.get("filters")
        self._distinct = parameters.get("distinct", False)

        # handle variation in how the unnested column is represented
        if self._unnest_column.node_type == NodeType.NESTED:
            self._unnest_column = self._unnest_column.centre

        # if we have a literal that's not a tuple, wrap it
        if self._unnest_column.node_type == NodeType.LITERAL and not isinstance(
            self._unnest_column.value, tuple
        ):
            self._unnest_column.value = tuple([self._unnest_column.value])

        self._single_column = parameters.get("pre_update_columns", set()) == {
            self._unnest_target.identity,
        }

        self.hash_set = FlatHashSet()

    @property
    def name(self):  # pragma: no cover
        return "Cross Join"

    @property
    def config(self):  # pragma: no cover
        filters = ""
        if self._filters:
            filters = f"({self._unnest_target.name} IN ({', '.join(self._filters)}))"
        return f"CROSS JOIN {filters}"

    def execute(self, morsel: pyarrow.Table, join_leg: str = None) -> pyarrow.Table:
        morsel = self.ensure_arrow_table(morsel)

        if morsel == EOS:
            yield EOS
            return
        if isinstance(self._unnest_column.value, tuple):
            yield from _cross_join_unnest_literal(
                morsel=morsel,
                source=self._unnest_column.value,
                target_column=self._unnest_target,
            )
            return

        yield from _cross_join_unnest_column(
            morsel=morsel,
            source=self._unnest_column,
            target_column=self._unnest_target,
            conditions=self._filters,
            hash_set=self.hash_set,
            distinct=self._distinct,
            single_column=self._single_column,
        )
        return



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

import fnmatch
from typing import Dict
from typing import Iterable
from typing import List

import orjson

from opteryx.config import RESOURCES_PATH


def load_permissions() -> List[Dict]:
    """
    Load permissions from a JSON file and return a list of permissions.

    If the file is not found, or an error occurs, return a default set of permissions.

    Returns:
        List[Dict]: A list of dictionaries where each dictionary represents a permission.
    """
    try:
        with open(RESOURCES_PATH / "permissions.json", "r", encoding="UTF8") as file:
            # Load each line as a JSON object and append a default permission entry.
            _permissions = [orjson.loads(line) for line in file] + [
                {"role": "opteryx", "permission": "READ", "table": "*"}
            ]
        return _permissions
    except FileNotFoundError:
        # Return a default permission if the file does not exist.
        return [{"role": "opteryx", "permission": "READ", "table": "*"}]
    except Exception as err:
        # Log the error and return a default permission in case of any other exceptions.
        print(f"[OPTERYX] Failed to load permissions: {err}")
        return [{"role": "opteryx", "permission": "READ", "table": "*"}]


# Load permissions once and make them globally accessible.
PERMISSIONS: List[Dict] = load_permissions()


def can_read_table(roles: Iterable[str], table: str, action: str = "READ") -> bool:
    """
    Check if any of the provided roles have READ access to the specified table.

    When we call this function, we provide the current user's roles and the table name.
    We then check if any of the permissions in the system match those roles and if those permissions
    grant access to the table.

    Tables can have wildcards in their names, so we use fnmatch to check the table name.

    We have a default role 'opteryx' with READ access to all tables.

    Parameters:
        roles (List[str]): A list of roles to check against permissions.
        table (str): The name of the table to check access for.

    Returns:
        bool: True if any role has READ access to the table, False otherwise.
    """

    def escape_special_chars(pattern: str) -> str:
        return pattern.replace(r"\*", "*").replace(r"\?", "?")

    # If no permissions are loaded, default to allowing all reads.
    if not PERMISSIONS:
        return True

    table = escape_special_chars(table)

    for entry in PERMISSIONS:
        # Check if the permission, the role is in the provided roles,
        # and the table matches the pattern defined in the permission.
        if (
            entry["permission"] == action
            and entry["role"] in roles
            and fnmatch.fnmatch(table, entry["table"])
        ):
            # Additional check for leading dots
            if table.startswith(".") and not entry["table"].startswith("."):
                continue
            return True

    # If no matching permission is found, deny access.
    return False



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import numpy
cimport numpy
numpy.import_array()

from libc.time cimport time

cdef bytes alphabet = b"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_/"

# Seed for xorshift32 PRNG
cdef unsigned int xorshift32_state = <unsigned int>time(NULL)


def list_random_strings(int row_count, int width) -> numpy.ndarray:
    """
    Generates a NumPy array of random fixed-width strings, repeated `row_count` times.

    Parameters:
        row_count: int
            The number of random strings to generate.
        width: int
            The length of each random string.

    Returns:
        A NumPy array containing `row_count` random strings of fixed width.
    """

    # Allocate NumPy array with fixed-width strings, dtype='S{width}'
    cdef numpy.ndarray result = numpy.empty((row_count,), dtype=f'S{width}')

    cdef unsigned int total_chars = row_count * width
    cdef unsigned int i
    cdef unsigned char rand_value
    cdef char* ptr = <char*>result.data

    for i from 0 <= i < total_chars:
        rand_value = xorshift32() & 0x3F  # Random value limited to 64 (alphabet size)
        ptr[0] = alphabet[rand_value]
        ptr += 1

    return result


cdef inline unsigned int xorshift32():
    global xorshift32_state  # Declare as global to modify the module-level variable
    cdef unsigned int x = xorshift32_state
    x ^= (x << 13)
    x ^= (x >> 17)
    x ^= (x << 5)
    xorshift32_state = x
    return x



# distutils: language = c++
# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=True
# cython: boundscheck=False

from libc.stdint cimport uint8_t, int64_t, uint64_t
from libc.math cimport isnan
from libcpp.unordered_map cimport unordered_map
from cython.operator cimport dereference, preincrement
from libcpp.string cimport string
from libc.string cimport memcpy, memset
from cpython.bytes cimport PyBytes_AsStringAndSize
from cpython.unicode cimport PyUnicode_AsUTF8AndSize, PyUnicode_Check

import datetime
from decimal import Decimal
from cython cimport cast

cdef uint64_t _INT64_HIGH_BIT = (<uint64_t>1) << 63
cdef int64_t NULL_FLAG = -cast(int64_t, _INT64_HIGH_BIT)              # -9223372036854775808
cdef int64_t MIN_SIGNED_64BIT = NULL_FLAG + 1               # -9223372036854775807
cdef int64_t MAX_SIGNED_64BIT = cast(int64_t, _INT64_HIGH_BIT - 1)    # 9223372036854775807

cdef inline bint map_contains(unordered_map[string, int64_t]& m, string key):
    return m.find(key) != m.end()

cdef inline int64_t map_get(unordered_map[string, int64_t]& m, string key, int64_t default_val=0):
    cdef unordered_map[string, int64_t].iterator it = m.find(key)
    if it != m.end():
        return m[key]
    return default_val

cdef inline int64_t _ensure_64bit_range(object val):
    if val < MIN_SIGNED_64BIT:
        return MIN_SIGNED_64BIT
    if val > MAX_SIGNED_64BIT:
        return MAX_SIGNED_64BIT
    return val

cdef inline int64_t decode_signed_big_endian_bytes(const char* buf):
    cdef uint64_t result = 0
    cdef int i
    for i in range(8):
        result = (result << 8) | <uint8_t>buf[i]
    return <int64_t>result

cpdef int64_t to_int(object value):
    """
    Convert a value to a signed 64-bit int for order-preserving comparisons.
    Returns MIN_SIGNED_64BIT to MAX_SIGNED_64BIT or raises ValueError.
    """
    cdef int64_t result
    cdef type value_type = type(value)

    cdef char* raw
    cdef const char* _raw
    cdef Py_ssize_t length
    cdef char buf[8]

    if value_type == int:
        return _ensure_64bit_range(value)

    if value_type == float:
        if value == float("inf"):
            return MAX_SIGNED_64BIT
        if value == float("-inf"):
            return MIN_SIGNED_64BIT
        if isnan(value):
            return NULL_FLAG
        return _ensure_64bit_range(<int64_t>round(value))

    if value_type == datetime.datetime:
        return _ensure_64bit_range(<int64_t>round(value.timestamp()))

    if value_type == datetime.date:
        # Converts to days since epoch (1970-01-01)
        timestamp = int(value.strftime("%s"))
        return _ensure_64bit_range(<int64_t>timestamp)

    if value_type == datetime.time:
        result = value.hour * 3600 + value.minute * 60 + value.second
        return _ensure_64bit_range(result)

    if value_type == Decimal:
        return _ensure_64bit_range(<int64_t>round(value))

    if value_type == str:
        # Keep the first byte as 0 to ensure the order of the 64bit int
        # is preserved by creating negative numbers
        memset(buf, 0, 8)

        if PyUnicode_Check(value):
            _raw = PyUnicode_AsUTF8AndSize(value, &length)
            memcpy(&buf[1], _raw, min(length, 7))
        else:
            PyBytes_AsStringAndSize(value, &raw, &length)
            memcpy(&buf[1], raw, min(length, 7))

        return _ensure_64bit_range(decode_signed_big_endian_bytes(buf))

    if value_type == bytes:
        # Keep the first byte as 0 to ensure order is preserved
        memset(buf, 0, 8)
        PyBytes_AsStringAndSize(value, &raw, &length)
        memcpy(&buf[1], raw, min(length, 7))
        return _ensure_64bit_range(decode_signed_big_endian_bytes(buf))

    return NULL_FLAG

cdef inline void write_int64(char* buf, int offset, int64_t value):
    cdef int i
    for i in range(8):
        buf[offset + 7 - i] = <char>((value >> (i * 8)) & 0xFF)

cdef inline int64_t read_int64(const char* buf, int offset):
    cdef int64_t result = 0
    cdef int i
    for i in range(8):
        result = (result << 8) | (<unsigned char>buf[offset + i])
    return result


cdef inline void write_map(unordered_map[string, int64_t]& m, bytearray out):

    cdef int64_t size = <int64_t>m.size()
    cdef unordered_map[string, int64_t].iterator it = m.begin()
    cdef string key
    cdef int64_t value

    out += (<int64_t>size).to_bytes(4, "big")

    # Iterate through all entries
    while it != m.end():
        # Access current key-value pair
        key = dereference(it).first
        value = dereference(it).second

        out.append(len(key))
        out += key
        out += (<int64_t>value).to_bytes(8, "big", signed=True)

        preincrement(it)


cdef inline void read_map(const char* buf, Py_ssize_t* offset, unordered_map[string, int64_t]& m):
    cdef int64_t size = 0
    cdef int i

    for i in range(4):
        size = (size << 8) | (<unsigned char>buf[offset[0] + i])
    offset[0] = offset[0] + 4
    for i in range(size):
        key_len = <unsigned char>buf[offset[0]]
        offset[0] = offset[0] + 1
        key = string(buf + offset[0], key_len)
        offset[0] = offset[0] + key_len
        val = read_int64(buf, offset[0])
        offset[0] = offset[0] + 8
        m[key] = val

cdef class RelationStatistics:
    cdef public int64_t record_count
    cdef public int64_t record_count_estimate
    cdef public unordered_map[string, int64_t] null_count
    cdef public unordered_map[string, int64_t] lower_bounds
    cdef public unordered_map[string, int64_t] upper_bounds
    cdef public unordered_map[string, int64_t] cardinality_estimate

    def __cinit__(self):
        self.record_count = 0
        self.record_count_estimate = 0
        self.null_count = {}
        self.lower_bounds = {}
        self.upper_bounds = {}
        self.cardinality_estimate = {}

    cpdef void update_lower(self, str column, object value, object index=None):
        cdef int64_t int_value = to_int(value)
        cdef string skey = column.encode()

        if int_value != NULL_FLAG:
            if not map_contains(self.lower_bounds, skey) or int_value < self.lower_bounds[skey]:
                self.lower_bounds[skey] = int_value

    cpdef void update_upper(self, str column, object value, object index=None):
        cdef int64_t int_value = to_int(value)
        cdef string skey = column.encode()

        if int_value != NULL_FLAG:
            if not map_contains(self.upper_bounds, skey) or int_value > self.upper_bounds[skey]:
                self.upper_bounds[skey] = int_value

    cpdef void add_null(self, str column, int nulls):
        cdef string skey = column.encode()

        self.null_count[skey] = map_get(self.null_count, skey, 0) + nulls

    cpdef void set_cardinality_estimate(self, str column, int cardinality):
        cdef string skey = column.encode()

        self.cardinality_estimate[skey] = cardinality

    cpdef bytes to_bytes(self):
        cdef bytearray out = bytearray()
        out += (<int64_t>self.record_count).to_bytes(8, "big")
        out += (<int64_t>self.record_count_estimate).to_bytes(8, "big")
        write_map(self.null_count, out)
        write_map(self.lower_bounds, out)
        write_map(self.upper_bounds, out)
        write_map(self.cardinality_estimate, out)
        return bytes(out)

    @classmethod
    def from_bytes(cls, bytes data):
        cdef const char* buf
        buf = data
        cdef Py_ssize_t offset = 0
        cdef RelationStatistics inst = cls()
        inst.record_count = read_int64(buf, offset)
        offset += 8
        inst.record_count_estimate = read_int64(buf, offset)
        offset += 8
        read_map(buf, &offset, inst.null_count)
        read_map(buf, &offset, inst.lower_bounds)
        read_map(buf, &offset, inst.upper_bounds)
        read_map(buf, &offset, inst.cardinality_estimate)
        return inst

    cpdef void merge(self, RelationStatistics other):
        """
        Merge another RelationStatistics into this one.

        Parameters:
            other: RelationStatistics
                The statistics object to merge into this one.

        Behavior:
            - Sums record counts
            - Takes minimum of lower bounds
            - Takes maximum of upper bounds
            - Sums null counts
            - Takes maximum of cardinality estimates
        """
        cdef unordered_map[string, int64_t].iterator it
        cdef string key
        cdef int64_t value

        # Add record counts
        self.record_count += other.record_count
        self.record_count_estimate += other.record_count_estimate

        # Merge lower bounds (take minimum)
        it = other.lower_bounds.begin()
        while it != other.lower_bounds.end():
            key = dereference(it).first
            value = dereference(it).second
            if not map_contains(self.lower_bounds, key) or value < self.lower_bounds[key]:
                self.lower_bounds[key] = value
            preincrement(it)

        # Merge upper bounds (take maximum)
        it = other.upper_bounds.begin()
        while it != other.upper_bounds.end():
            key = dereference(it).first
            value = dereference(it).second
            if not map_contains(self.upper_bounds, key) or value > self.upper_bounds[key]:
                self.upper_bounds[key] = value
            preincrement(it)

        # Merge null counts (sum)
        it = other.null_count.begin()
        while it != other.null_count.end():
            key = dereference(it).first
            value = dereference(it).second
            self.null_count[key] = map_get(self.null_count, key, 0) + value
            preincrement(it)

        # Merge cardinality estimates (take maximum)
        it = other.cardinality_estimate.begin()
        while it != other.cardinality_estimate.end():
            key = dereference(it).first
            value = dereference(it).second
            if not map_contains(self.cardinality_estimate, key) or value > self.cardinality_estimate[key]:
                self.cardinality_estimate[key] = value
            preincrement(it)

    def __deepcopy__(self, memo):
        return self
